"""
layer.py - Layer of content drawn on a Tablet
"""
# System
import logging
from typing import TYPE_CHECKING, List

# Qt
from PyQt6.QtWidgets import (QGraphicsRectItem, QGraphicsLineItem, QGraphicsPolygonItem,
                             QGraphicsItemGroup)

# Tablet
import tabletqt.element as element
from tabletqt.presentation import Presentation
from tabletqt.graphics.circle_se import CircleSE
from tabletqt.graphics.polygon_se import PolygonSE
from tabletqt.graphics.line_segment import LineSegment
from tabletqt.graphics.rectangle_se import RectangleSE
from tabletqt.graphics.text_element import TextElement
from tabletqt.graphics.image import ImageDE
from tabletqt.graphics.symbol import Symbol
from tabletqt.graphics.diagnostic_marker import DiagnosticMarker


if TYPE_CHECKING:
    from tabletqt.tablet import Tablet

class Layer:
    """
    A common feature of many drawing and image editing applications is the ability to stack layers of
    content along a z axis toward the user’s viewpoint. Similarly, the Tablet renders each layer working
    from the lowest value on the z axis toward the highest. Thus, content on the higher layers may overlap
    content underneath.

        Attributes and relationships defined on the class model

        - Name {I} -- A name that is unique among other Layers on this Tablet (attribute)
        - Z coord {I2, OR20} -- Rendering order, also unique on this Tablet (implement as Tablet list)
        - Presentation {R2} -- Defines styling of text and graphics (attribute)
        - Drawing type {R?} -- Determines assets that can be drawn (attribute)
        - Tablet {R13} -- Layer is managed on this Tablet (attribute)
    """

    def __init__(self, name: str, tablet: 'Tablet', presentation: str, drawing_type: str):
        """
        Constructor

        :param name: A predefined or custom Layer name
        :param tablet: Layer is created on this Tablet object
        :param presentation: Name of Presentation styling the text and graphics on this Layer
        :param drawing_type: Name of Drawing Type to determine assets that may be drawn on this Layer
        """
        self.logger = logging.getLogger(__name__)
        self.logger.info(f"creating layer: [{name}] ")
        self.Name = name
        self.Tablet = tablet
        self.Scene = tablet.View.scene  # This is what we actually draw on
        self.Drawing_type = drawing_type

        # Stuff we will draw on the Layer
        self.Line_segments: List[element.Line_Segment] = []
        self.Symbols: List[QGraphicsItemGroup] = []
        self.Circles: List[element.Circle] = []
        self.Polygons: List[QGraphicsPolygonItem] = []
        self.Rectangles: List[element.Rectangle] = []
        self.RawRectangles: List[QGraphicsRectItem] = []
        self.RawLines: List[QGraphicsLineItem] = []
        self.TextUnderlayRects: List[element.FillRect] = []
        self.Text: List[element.Text_line] = []
        self.Images: List[element.Image] = []

        # Load this Layer's presentation assets if they haven't been already
        # Unique ID (see Tablet Subsystem class diagram) of a Presentation is both
        # its name and its View Type name.  So we combine them to form the index
        pres_index = ':'.join([self.Drawing_type, presentation])
        self.Presentation = self.Tablet.Presentations.get(pres_index)
        if not self.Presentation:
            # It hasn't been loaded from the Flatland DB yet
            self.Presentation = Presentation(name=presentation, drawing_type=self.Drawing_type)
            self.Tablet.Presentations[pres_index] = self.Presentation


    def render(self):
        """Renders all Elements on this Layer"""

        self.logger.info(f'Rendering layer: {self.Name}')

        # Rendering order determines what can potentially overlap on this Layer, so order matters
        LineSegment.render(self)
        Symbol.render(self)
        CircleSE.render(self)
        RectangleSE.render(self)
        PolygonSE.render(self)
        TextElement.render_underlays(self)  # Renders any color fills that lie underneath text blocks or lines
        TextElement.render(self)  # Render text after vector content so that it is never underneath
        ImageDE.render(self)  # Text should not be drawn over images, so we can render these last

        # Diagnostic elements with explicit styling that bypass StyleDB lookup
        # Not intended for use by any client applications
        DiagnosticMarker.render(self)



""" test_state_symbols.py - Draw all of the state machine symbols for different presentation """

import pytest
import inspect
from pathlib import Path
from tabletqt.tablet import Tablet
from tabletqt.geometry_types import Rect_Size, Position
from tabletqt.graphics.symbol import Symbol

points_in_mm = 2.83465
A4 = Rect_Size(round(210 * points_in_mm), round(297 * points_in_mm))

def test_state_symbols():
    dtype = "xUML state machine diagram"
    function_name = inspect.currentframe().f_code.co_name
    output_path = Path(f"output/{function_name}.pdf")

    test_tablet = Tablet(size=A4, output_file=output_path, drawing_type=dtype,
                         presentation="default", layer="diagram", show_window=False, background_color='autumn')
    dlayer = test_tablet.layers['diagram']

    s = Symbol(layer=dlayer, name='target state', pin=Position(100, 200), angle=0)
    s = Symbol(layer=dlayer, name='initial pseudo state', pin=Position(200, 200), angle=0)
    s = Symbol(layer=dlayer, name='final pseudo state', pin=Position(300, 200), angle=0)

    s = Symbol(layer=dlayer, name='target state', pin=Position(100, 400), angle=90)
    s = Symbol(layer=dlayer, name='initial pseudo state', pin=Position(200, 400), angle=180)
    s = Symbol(layer=dlayer, name='final pseudo state', pin=Position(300, 400), angle=270)

    test_tablet.render()
    assert True

def test_bp_state_symbols():
    dtype = "xUML state machine diagram"
    function_name = inspect.currentframe().f_code.co_name
    output_path = Path(f"output/{function_name}.pdf")

    test_tablet = Tablet(size=A4, output_file=output_path, drawing_type=dtype,
                         presentation="blueprint", layer="diagram", show_window=False, background_color='blueprint')
    dlayer = test_tablet.layers['diagram']

    s = Symbol(layer=dlayer, name='target state', pin=Position(100, 200), angle=0)
    s = Symbol(layer=dlayer, name='initial pseudo state', pin=Position(200, 200), angle=0)
    s = Symbol(layer=dlayer, name='final pseudo state', pin=Position(300, 200), angle=0)

    s = Symbol(layer=dlayer, name='target state', pin=Position(100, 400), angle=90)
    s = Symbol(layer=dlayer, name='initial pseudo state', pin=Position(200, 400), angle=180)
    s = Symbol(layer=dlayer, name='final pseudo state', pin=Position(300, 400), angle=270)

    test_tablet.render()
    assert True




"""
exceptions.py – Tabletx exceptions
"""

class TabletException(Exception):
    pass

class BadConfigData(TabletException):
    pass

class MissingConfigData(TabletException):
    pass

class NonSystemInitialLayer(TabletException):
    pass

class TabletBoundsExceeded(TabletException):
    def __init__(self, message, height, width):
        self.message = message
        self.height = height
        self.width = width
        super().__init__(f"{message} (Height exceeded: {height} Width exceeded: {width})")




""" test_xuml_stickers.py - Draw all of the xuml class diagram stickers for different presentations """

import pytest
import inspect
from pathlib import Path
from tabletqt.tablet import Tablet
from tabletqt.geometry_types import Rect_Size, Position
from tabletqt.graphics.symbol import Symbol
from tabletqt.graphics.text_element import TextElement, TextBlockCorner, HorizAlign

points_in_mm = 2.83465
A4 = Rect_Size(round(210 * points_in_mm), round(297 * points_in_mm))

def test_xuml_stickers():
    dtype = "xUML class diagram"
    function_name = inspect.currentframe().f_code.co_name
    output_path = Path(f"output/{function_name}.pdf")

    test_tablet = Tablet(size=A4, output_file=output_path, drawing_type=dtype,
                         presentation="default", layer="diagram", show_window=False, background_color='blue steel')
    dlayer = test_tablet.layers['diagram']

    TextElement.add_sticker(layer=dlayer, asset='class face name', name='1 mult',
                            pin=Position(150, 400), corner=TextBlockCorner.LL)
    TextElement.add_sticker(layer=dlayer, asset='class face name', name='M mult',
                            pin=Position(175, 400), corner=TextBlockCorner.LL)
    TextElement.add_sticker(layer=dlayer, asset='class face name', name='1c mult',
                            pin=Position(200, 400), corner=TextBlockCorner.LL)
    TextElement.add_sticker(layer=dlayer, asset='class face name', name='Mc mult',
                            pin=Position(250, 400), corner=TextBlockCorner.LL)
    TextElement.add_sticker(layer=dlayer, asset='association name', name='M mult',
                            pin=Position(300, 400), corner=TextBlockCorner.LL)
    TextElement.add_sticker(layer=dlayer, asset='superclass face name', name='superclass',
                            pin=Position(350, 400), corner=TextBlockCorner.LL)

    test_tablet.render()
    assert True

def test_bp_xuml_stickers():
    dtype = "xUML class diagram"
    function_name = inspect.currentframe().f_code.co_name
    output_path = Path(f"output/{function_name}.pdf")

    test_tablet = Tablet(size=A4, output_file=output_path, drawing_type=dtype,
                         presentation="blueprint", layer="diagram", show_window=False, background_color='blueprint')
    dlayer = test_tablet.layers['diagram']

    TextElement.add_sticker(layer=dlayer, asset='class face name', name='1 mult',
                            pin=Position(150, 400), corner=TextBlockCorner.LL)
    TextElement.add_sticker(layer=dlayer, asset='class face name', name='M mult',
                            pin=Position(175, 400), corner=TextBlockCorner.LL)
    TextElement.add_sticker(layer=dlayer, asset='class face name', name='1c mult',
                            pin=Position(200, 400), corner=TextBlockCorner.LL)
    TextElement.add_sticker(layer=dlayer, asset='class face name', name='Mc mult',
                            pin=Position(250, 400), corner=TextBlockCorner.LL)
    TextElement.add_sticker(layer=dlayer, asset='association name', name='M mult',
                            pin=Position(300, 400), corner=TextBlockCorner.LL)
    TextElement.add_sticker(layer=dlayer, asset='superclass face name', name='superclass',
                            pin=Position(350, 400), corner=TextBlockCorner.LL)

    test_tablet.render()
    assert True




""" test_cd_text.py - Draw class diagram text blocks """

import inspect
from pathlib import Path
from tabletqt.tablet import Tablet
from tabletqt.geometry_types import Rect_Size, Position
from tabletqt.graphics.text_element import TextElement, TextBlockCorner, HorizAlign
from tabletqt.graphics.line_segment import LineSegment

points_in_mm = 2.83465
A4 = Rect_Size(round(210 * points_in_mm), round(297 * points_in_mm))

def test_cd_text():
    dtype = "Starr class diagram"
    function_name = inspect.currentframe().f_code.co_name
    output_path = Path(f"output/{function_name}.pdf")

    test_tablet = Tablet(size=A4, output_file=output_path, drawing_type=dtype,
                         presentation="default", layer="diagram", show_window=False, background_color='blue steel')
    dlayer = test_tablet.layers['diagram']

    vphrase = ["draws", "boundary with"]

    TextElement.pin_block(layer=dlayer, asset='class face name', pin=Position(200,300), text=vphrase,
                          corner=TextBlockCorner.UR, align=HorizAlign.RIGHT)

    TextElement.pin_block(layer=dlayer, asset='class face name', pin=Position(300,300), text=vphrase,
                          corner=TextBlockCorner.UR, align=HorizAlign.LEFT)
    test_tablet.render()
    assert True

def test_bp_cd_text():
    dtype = "Starr class diagram"
    function_name = inspect.currentframe().f_code.co_name
    output_path = Path(f"output/{function_name}.pdf")

    test_tablet = Tablet(size=A4, output_file=output_path, drawing_type=dtype,
                         presentation="blueprint", layer="diagram", show_window=False, background_color='blueprint')
    dlayer = test_tablet.layers['diagram']

    vphrase = ["draws", "boundary with"]

    TextElement.pin_block(layer=dlayer, asset='class face name', pin=Position(200,300), text=vphrase,
                          corner=TextBlockCorner.UR, align=HorizAlign.RIGHT)

    TextElement.pin_block(layer=dlayer, asset='class face name', pin=Position(300,300), text=vphrase,
                          corner=TextBlockCorner.UR, align=HorizAlign.LEFT)
    test_tablet.render()
    assert True

def test_Starr_underlay():
    dtype = "Starr class diagram"
    function_name = inspect.currentframe().f_code.co_name
    output_path = Path(f"output/{function_name}.pdf")

    test_tablet = Tablet(size=A4, output_file=output_path, drawing_type=dtype,
                         presentation="default", layer="diagram", show_window=False, background_color='blue steel')
    dlayer = test_tablet.layers['diagram']

    LineSegment.add(layer=dlayer, asset="binary association connector",
                    from_here=Position(100, 305), to_there=Position(400, 305))
    TextElement.add_block(layer=dlayer, asset='generalization', lower_left=Position(300, 300), text=['underlay'])

    vphrase = ["draws", "boundary with"]

    TextElement.pin_block(layer=dlayer, asset='class face name', pin=Position(200,300), text=vphrase,
                          corner=TextBlockCorner.UR, align=HorizAlign.RIGHT)

    TextElement.pin_block(layer=dlayer, asset='class face name', pin=Position(300,300), text=vphrase,
                          corner=TextBlockCorner.UR, align=HorizAlign.LEFT)
    test_tablet.render()
    assert True

def test_xUML_underlay():
    dtype = "xUML class diagram"
    function_name = inspect.currentframe().f_code.co_name
    output_path = Path(f"output/{function_name}.pdf")

    test_tablet = Tablet(size=A4, output_file=output_path, drawing_type=dtype,
                         presentation="default", layer="diagram", show_window=False, background_color='blue steel')
    dlayer = test_tablet.layers['diagram']

    LineSegment.add(layer=dlayer, asset="binary association connector",
                    from_here=Position(100, 305), to_there=Position(400, 305))

    TextElement.add_sticker(layer=dlayer, asset='superclass face name', name='superclass',
                            pin=Position(300, 300), corner=TextBlockCorner.LL)

    test_tablet.render()
    assert True



""" image.py -- Image element """

# System
import logging
from pathlib import Path
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from tabletqt.layer import Layer

# Qt
from PyQt6.QtWidgets import QGraphicsPixmapItem
from PyQt6.QtGui import QPixmap

# Modelint
from mi_config.config import Config

# Tablet
import tabletqt.element as element
from tabletqt.geometry_types import Position, Rect_Size
from tabletqt.exceptions import TabletBoundsExceeded
from tabletqt.styledb import StyleDB
from tabletqt.tablet_config import TabletConfig

_logger = logging.getLogger(__name__)

class ImageDE:
    """
    Image Display Element
    (subclass of Rectangular Display Element in Render Subsystem not shown on model yet)

    Manage the loading and rendering of Image Elements

    Attributes and relationships defined on the class model

    (Not currently in the class model, but will be added)

    Subclass of Element on class model (R15) - most likely

    - ID {I} -- Element ID, unique within a Layer, implemented as object reference
    - Layer {I, R22} -- Element drawn on this Layer via R22/R12/R15/Element/R19/Layer
    - Location -- Path to the image file
    - Lower left -- Position of the lower left image corner in tablet coordinates
    - Size -- The actual size of the image in the file
    """
    image_paths = None

    @classmethod
    def build_paths(cls):
        """ Assign image file dictionary """
        raw_config_data = Config(app_name=TabletConfig.app_name, lib_config_dir=TabletConfig.config_path, fspec={'images': None})
        image_dict = raw_config_data.loaded_data['images']
        root_dir = raw_config_data.user_config_dir
        cls.image_paths = {k: Path(root_dir / 'images' / v) for k,v in image_dict.items()}


    @classmethod
    def add(cls, layer: 'Layer', name: str, lower_left: Position, size: Rect_Size):
        """
        Adds the image to the layer, converting to device coordinates and using the upper
        left corner of the image instead to suit Qt placement

        :param layer: Draw on this Layer
        :param name: Name of the image resource supplied by the client
        :param size: The actual size of the image in points
        :param lower_left:  Lower left corner of the image in tablet coordinates
        """

        # Flip lower left corner to device coordinates
        try:
            ll_dc = layer.Tablet.to_dc(Position(x=lower_left.x, y=lower_left.y))
        except TabletBoundsExceeded:
            _logger.warning(f"Lower left corner of image [{image_path}] is outside Tablet draw region")
            return

        # Use upper left corner instead
        ul = Position(x=ll_dc.x, y=ll_dc.y - size.height)

        # Add it to the list
        try:
            layer.Images.append(element.Image(resource_path=cls.image_paths[name], upper_left=ul, size=size))
        except KeyError:
            _logger.warning(f"No path defined for image [{name}]")
        else:
            _logger.info(f'View>> Layer {layer.Name} registered resource at: {cls.image_paths[name]}')

    @classmethod
    def render(cls, layer: 'Layer'):
        """
        Render the Image Elements

        :param layer: Draw on this Layer
        """
        for i in layer.Images:
            if not i.resource_path.exists():
                _logger.error(f'Image file [{i.resource_path}] not found')
                continue

            pixmap = QPixmap(str(i.resource_path))
            if pixmap.isNull():
                _logger.error(f'Image file [{i.resource_path}] could not be loaded as pixmap')
                continue

            pix_item = QGraphicsPixmapItem(pixmap)
            pix_item.setPos(i.upper_left.x, i.upper_left.y)
            layer.Scene.addItem(pix_item)
            _logger.info(f"> Image at: ({i.upper_left.x}, {i.upper_left.y})")



""" test_rectangles.py - Draw some rectangle shapes for different presentations on the diagram layer """

import pytest
import inspect
from pathlib import Path
from tabletqt.tablet import Tablet
from tabletqt.geometry_types import Rect_Size, Position
from tabletqt.graphics.rectangle_se import RectangleSE

points_in_mm = 2.83465
A4 = Rect_Size(round(210 * points_in_mm), round(297 * points_in_mm))

def test_grid_boundary():
    dtype = "Grid Diagnostic"
    function_name = inspect.currentframe().f_code.co_name
    output_path = Path(f"output/{function_name}.pdf")

    test_tablet = Tablet(size=A4, output_file=output_path, drawing_type=dtype,
                         presentation="default", layer="grid", show_window=False, background_color='white')
    glayer = test_tablet.layers['grid']

    r = RectangleSE.add(layer=glayer, asset="grid boundary", lower_left=Position(10, 10),
                        size=Rect_Size(height=400, width=400))
    test_tablet.render()
    assert True

def test_class_compartment():
    dtype = "Starr class diagram"
    function_name = inspect.currentframe().f_code.co_name
    output_path = Path(f"output/{function_name}.pdf")

    test_tablet = Tablet(size=A4, output_file=output_path, drawing_type=dtype,
                         presentation="default", layer="diagram", show_window=False, background_color='blue steel')
    dlayer = test_tablet.layers['diagram']

    r = RectangleSE.add(layer=dlayer, asset="class name compartment", lower_left=Position(100, 100),
                        size=Rect_Size(height=27, width=253))
    test_tablet.render()
    assert True

def test_bp_class_compartment():
    dtype = "Starr class diagram"
    function_name = inspect.currentframe().f_code.co_name
    output_path = Path(f"output/{function_name}.pdf")

    test_tablet = Tablet(size=A4, output_file=output_path, drawing_type=dtype,
                         presentation="blueprint", layer="diagram", show_window=False, background_color='blueprint')
    dlayer = test_tablet.layers['diagram']

    r = RectangleSE.add(layer=dlayer, asset="imported class attribute compartment", lower_left=Position(100, 100),
                        size=Rect_Size(height=27, width=253))
    test_tablet.render()
    assert True

def test_state_compartment():
    dtype = "xUML state machine diagram"
    function_name = inspect.currentframe().f_code.co_name
    output_path = Path(f"output/{function_name}.pdf")

    test_tablet = Tablet(size=A4, output_file=output_path, drawing_type=dtype,
                         presentation="default", layer="diagram", show_window=False, background_color='autumn')
    dlayer = test_tablet.layers['diagram']

    r = RectangleSE.add(layer=dlayer, asset="state name only name compartment", lower_left=Position(100, 100),
                        size=Rect_Size(height=27, width=253))
    test_tablet.render()
    assert True

def test_bp_state_compartment():
    dtype = "xUML state machine diagram"
    function_name = inspect.currentframe().f_code.co_name
    output_path = Path(f"output/{function_name}.pdf")

    test_tablet = Tablet(size=A4, output_file=output_path, drawing_type=dtype,
                         presentation="blueprint", layer="diagram", show_window=False, background_color='blueprint')
    dlayer = test_tablet.layers['diagram']

    r = RectangleSE.add(layer=dlayer, asset="state name only compartment", lower_left=Position(100, 100),
                        size=Rect_Size(height=27, width=253))
    test_tablet.render()
    assert True



""" tablet_config.py -- Configuration class from Render Subsystem class model """

# System
from pathlib import Path

class TabletConfig:
    """

    """
    config_path = Path(__file__).parent / "configuration"
    image_path = config_path / "images"
    app_name = "mi_tablet"



""" demo6_bp_rects.py - Test rectangles using blueprint presentation """

from pathlib import Path
from tabletqt.tablet import Tablet
from tabletqt.geometry_types import Rect_Size, Position
from tabletqt.graphics.symbol import Symbol
from tabletqt.graphics.text_element import TextElement, TextBlockCorner, HorizAlign
from tabletqt.graphics.rectangle_se import RectangleSE
from tabletqt.graphics.diagnostic_marker import DiagnosticMarker

points_in_mm = 2.83465
points_in_cm = 28.3465
points_in_inch = 72

class SketchSymbols:
    tablet = None
    layer = None
    size = None

    @classmethod
    def make_a_tablet(cls):
        cls.size = Rect_Size(round(210*points_in_mm), round(297*points_in_mm))  # Ansi C portrait
        dtype = "Starr class diagram"
        # dtype = "xUML class diagram"
        # dtype = "xUML state machine diagram"
        output_path = Path(__file__).parent.parent.parent.parent / "working" / "demo6_bp_rects.pdf"
        cls.tablet = Tablet(size=cls.size, output_file=output_path, drawing_type=dtype,
                            presentation="blueprint", layer="diagram", show_window=False, background_color='blueprint')

    @classmethod
    def sketch(cls):

        cls.make_a_tablet()
        dlayer = cls.tablet.layers['diagram']

        r = RectangleSE.add(layer=dlayer, asset="imported class attribute compartment", lower_left=Position(100, 100),
                            size=Rect_Size(height=27, width=253))
        cls.tablet.render()

if __name__ == "__main__":
    SketchSymbols.sketch()



""" crayon_box.py - Sets the Qt pen and brush properties using the StyleDB """

# System
import logging
from typing import Optional

# Qt
from PyQt6.QtGui import QBrush, QPen, QColor
from PyQt6.QtCore import Qt
from PyQt6.QtWidgets import QAbstractGraphicsShapeItem, QGraphicsLineItem

# Tablet
from tabletqt.styledb import StyleDB, FloatRGB

_logger = logging.getLogger(__name__)
class CrayonBox:
    """
    Get your crayons here. Specifically, your Qt pen and brush settings.
    """
    @classmethod
    def choose_crayons(cls, item: QAbstractGraphicsShapeItem | QGraphicsLineItem, border_style: str,
                       fill: Optional['str'] = None):
        """
        Create and assign a pen and brush for a given QT graphics item with settings found
        in the StyleDB

        :param item: Qt graphic item to be displayed
        :param border_style: Name of the user border style key in the StyleDB
        :param fill: Name of user fill key in the StyleDB, if any
        """
        # Create a pen and set its properties using the StyleDB
        # Color
        cname = StyleDB.line_style[border_style].color
        c_color = StyleDB.color[cname]
        # Width
        w = StyleDB.line_style[border_style].width
        pen = QPen(QColor(*c_color), w)
        # Pattern
        pname = StyleDB.line_style[border_style].pattern  # name of line style's pattern
        if pname != 'no dash':
            pvalue = StyleDB.dash_pattern[pname]  # find pattern value in dash pattern dict
            pen.setDashPattern([*pvalue])

        if isinstance(item, QAbstractGraphicsShapeItem):
            # item is fillable, not a line item, for example
            if fill:
                # If a fill is specified, create a corresponding brush
                fill_rgb_color_value = StyleDB.color[fill]
                brush = QBrush(QColor(*fill_rgb_color_value))
                _logger.info(f"Brush color: [{fill}]")
            else:
                brush = QBrush(Qt.GlobalColor.transparent)
            item.setBrush(brush)

        # Set the item pen and brush of the graphic item
        item.setPen(pen)
        _logger.info(f"Pen color: [{cname}], width: [{w}], pattern: [{pname}]")

    @classmethod
    def choose_fill_only(cls, item: QAbstractGraphicsShapeItem, fill: FloatRGB):
        """
        Create and assign a transparent pen (no border) and brush for a given QT graphics item
        with settings found in the StyleDB.

        :param item: Qt graphic item to be displayed (must be a closed shape)
        :param fill: RGB color
        :return:
        """
        _logger.info(f"Brush color: {fill}")

        pen = QPen(Qt.GlobalColor.transparent)
        brush = QBrush(QColor(*fill))
        item.setPen(pen)
        item.setBrush(brush)


"""
element.py - Fundamental drawing elements such as shapes and text lines
"""
from collections import namedtuple

Circle = namedtuple('_Circle', 'center radius border_style fill')
"""Circle"""
Polygon = namedtuple('_Polygon', 'vertices border_style fill')
"""Closed polygon (other than a rectangle) that can be filled"""
Line_Segment = namedtuple('_Line_Segment', 'from_here to_there style')
"""A line segment is drawn from one point on the Tablet to another using some line drawing style"""
Rectangle = namedtuple('_Rectangle', 'upper_left size border_style fill radius top bottom')
"""A rectangle is positioned at its lower left corner and then drawn with the specified size"""
FillRect = namedtuple('_FillRect', 'upper_left size color')
"""A borderless rectangle is filled with the specified color"""
Text_line = namedtuple('_Text_line', 'upper_left text style')
"""A line of text (no CR/LF characters) rendered with a given style"""
Image = namedtuple('_Image', 'resource_path upper_left size')
"""A png or jpeg image file and a position"""



"""
geometry_types.py

Data types for 2D geometry
"""

from typing import NewType
from collections import namedtuple
from enum import Enum

Coordinate = NewType('Coordinate', float )
Distance = int


class VertAlign(Enum):
    """The numeric values are low to high in the axis direction"""
    BOTTOM = 0
    CENTER = 1
    TOP = 2


class HorizAlign(Enum):
    """The numeric values are low to high in the axis direction"""
    LEFT = 0
    CENTER = 1
    RIGHT = 2


Rectangle = namedtuple('Rectangle', 'line_style lower_left, size')
Position = namedtuple('Position', 'x y')
Line_Segment = namedtuple('Line_Segment', 'from_position to_position')
Rect_Size = namedtuple('Rect_Size', 'height width')


""" symbol.py - Draw a predefined symbol """

# System
import logging
from PyQt6.QtWidgets import QGraphicsPolygonItem, QGraphicsLineItem, QGraphicsItemGroup, QGraphicsEllipseItem
from PyQt6.QtCore import QPointF, QLineF, QRectF
from PyQt6.QtGui import QPolygonF
from typing import TYPE_CHECKING, Callable, Dict

if TYPE_CHECKING:
    from tabletqt.tablet import Layer

# Modelint
from mi_config.config import Config

# Tablet
from tabletqt.styledb import StyleDB
from tabletqt.geometry_types import Position
from tabletqt.graphics.crayon_box import CrayonBox
from tabletqt.tablet_config import TabletConfig
from tabletqt.exceptions import BadConfigData

class Symbol:
    """
    A composite group of shapes that can be rotated and placed anywhere on the Tablet on a specified Layer.
    """
    symbol_defs = None

    def __init__(self, layer: 'Layer', name: str, pin: Position, angle: int = 0):
        """
        Creates a Symbol that builds itself up from data specified in symbols.yaml and
            builds a PyQt group which is added to the layer's list of Symbols

        :param app: Name of the client application, ex: 'flatland', (needed to index into symbol data)
        :param layer: Layer where symbol will be drawn ex: 'diagram'
        :param group: Symbol belongs to this group (diagram type/notation if flatland), ex: Starr class
        :param name: Symbol name (1 mult, target state, etc)
        :param pin: Where the symbol will be 'pinned' in table coordinates. Imagine a pin pushed through the
        bottom of the symbol bounding box into the tablet. Then the symbol can be rotated around this pin. So
        if you are placing a double headed arrow with the arrow pointing upward in the symbol definition,
        for example, you pin the center base and then rotate the arrows leaving the base center in the same
        location regardless of rotation. In other words, the client need not compute the bounding box to determine
        the pin location.
        :param angle: Degrees clockwise with 0, 90, 180, and 270 at 12, 3, 6, and 9 o'clock respectively
        """
        self.logger = logging.getLogger(__name__)
        self.layer = layer
        self.name = name
        self.pin = pin
        self.angle = angle
        self.width = 0
        self.height = 0
        self.symbol_item = QGraphicsItemGroup()

        try:
            symbol_def = self.symbol_defs[self.layer.Drawing_type][name]
        except KeyError:
            self.logger.error(f"No symbol named [{name}] defined for drawing type [{self.layer.Drawing_type}]")
            raise BadConfigData

        self.component_methods: Dict[str, Callable[[dict], None]] = {
            'polygon': self.add_polygon,
            'polyline': self.add_polyline,
            'circle': self.add_circle,
        }

        # Convert pin to device coordinates
        self.device_pin = layer.Tablet.to_dc(pin)

        # Create item for each component shape element, position it on the tablet, and add it to a Qt
        # symbol item group for any transforms to be applied to the entire symbol
        for component_name, cdef in symbol_def.items():
            component_type = list(cdef.keys())[0]
            try:
                method = self.component_methods[component_type]
            except KeyError:
                self.logger.error(f"Component type {component_type} for symbol {name} not supported")
                raise BadConfigData
            method(component_name, cdef[component_type])  # Call method apporpriate to the shape

        # for shape in self.shape_elements:
        #     for sname, method in self.shape_methods.items():
        #         if shape.get(sname):
        #             method(shape)  # Call method apporpriate to the shape

        # Create the rotation transform and apply it to the group
        self.symbol_item.setTransformOriginPoint(self.device_pin.x, self.device_pin.y)
        self.symbol_item.setRotation(angle)

        # Add it the transformed group to the Symbols list in the specified layer of the tablet
        # for later rendering
        self.layer.Symbols.append(self.symbol_item)


    @classmethod
    def load_symbol_defs(cls):
        """
        Load all symbol definitions from the symbols.yaml file
        """
        raw_config_data = Config(app_name=TabletConfig.app_name, lib_config_dir=TabletConfig.config_path,
                                 fspec={'symbols': None})
        cls.symbol_defs = raw_config_data.loaded_data['symbols']

    def add_circle(self, component_name: str, shape_def):
        """
        Adds a circle shape element to the symbol item group

        :param component_name:
        :param shape_def: The shape definition obtained from the symbol yaml file for this circle
        """
        # Translate center in symbol bounding box relative to the pin
        # We just add the center position to the pin position to translate the circle
        center_tablet = Position(self.pin[0]+shape_def['center'][0], self.pin[1]+shape_def['center'][1])

        # Convert center tablet to device coordinates
        center_dc = self.layer.Tablet.to_dc(center_tablet)
        radius = shape_def['radius']
        diameter = radius * 2

        # Update dimensions of the bounding box if either dimension is larger
        self.width = max(self.width, diameter)
        self.height = max(self.height, diameter)

        # Create Qt circle item
        # We subtract the radius since Qt wants the top left corner for positioning
        circle_item = QGraphicsEllipseItem(QRectF(center_dc.x-radius, center_dc.y-radius, diameter, diameter))

        try:
            component_style = self.layer.Presentation.Symbol_presentation[self.name][component_name]
        except KeyError:
            self.logger.error(f"No style defined for component [{component_name}] in symbol [{self.name}]")
            raise BadConfigData

        # Set pen and brush (border/fill) styles
        CrayonBox.choose_crayons(
            item=circle_item,
            border_style=component_style['line style'],
            fill=component_style['fill'])

        # Add circle shape element to the symbol item group
        self.symbol_item.addToGroup(circle_item)

    def add_polyline(self, component_name: str, shape_def):
        """
        Adds a path element (series of connected line segments) to the symbol item group

        :param shape_def: The shape definition obtained from the symbol yaml file for this path
        """
        # Translate each vertex relative to the pin
        # We just add the pin to vertex coordinates
        canvas_vertices = [Position(v[0] + self.pin[0], v[1] + self.pin[1]) for v in shape_def]

        # Update dimensions of the bounding box if any vertex is outside the current dimensions
        max_x = max(v[0] for v in shape_def)
        max_y = max(v[1] for v in shape_def)
        self.width = max(self.width, max_x)
        self.height = max(self.height, max_y)

        # Convert vertices to device coordinates
        vertices_dc = [self.layer.Tablet.to_dc(v) for v in canvas_vertices]

        try:
            component_style = self.layer.Presentation.Symbol_presentation[self.name][component_name]
        except KeyError:
            self.logger.error(f"No style defined for component [{component_name}] in symbol [{self.name}]")
            raise BadConfigData

        # Create a line item connecting gap between each pair of vertices
        # from beginning of vertex list to the last vertex in the list
        # and add the line item to the symbol item group
        for v in range(len(vertices_dc) - 1):
            start_point = vertices_dc[v]
            end_point = vertices_dc[v + 1]

            # Create a line item connecting the points
            line = QGraphicsLineItem(QLineF(*start_point, *end_point))

            # Set line style
            CrayonBox.choose_crayons(
                item=line,
                border_style=component_style['line style'])

            # Add line to the symbol item group
            self.symbol_item.addToGroup(line)

    def add_polygon(self, component_name: str, shape_def):
        """
        Adds a polygon component to the symbol

        :param polyline:
        :param component_name:
        :param shape_def: The shape definition obtained from the symbol yaml file for this triangle
        """
        # Translate each vertex relative to the pin
        # We just add the pin to vertex coordinates
        canvas_vertices = [Position(v[0] + self.pin[0], v[1] + self.pin[1]) for v in shape_def]

        # Update dimensions of the bounding box if any vertex is outside the current dimensions
        max_x = max(v[0] for v in shape_def)
        max_y = max(v[1] for v in shape_def)
        self.width = max(self.width, max_x)
        self.height = max(self.height, max_y)
        #
        # # Convert vertices to device coordinates
        vertices_dc = [self.layer.Tablet.to_dc(v) for v in canvas_vertices]
        #
        # # Create a polygon item
        pverts = [QPointF(*v) for v in vertices_dc]
        polygon = QPolygonF(pverts)
        poly_item = QGraphicsPolygonItem(polygon)
        #
        try:
            component_style = self.layer.Presentation.Symbol_presentation[self.name][component_name]
        except KeyError:
            self.logger.error(f"No style defined for component [{component_name}] in symbol [{self.name}]")
            raise BadConfigData
        # # Set pen and brush (border/fill) styles
        CrayonBox.choose_crayons(
            item=poly_item,
            border_style=component_style['line style'],
            fill=component_style['fill'])

        # Add triangle to the symbol item group
        self.symbol_item.addToGroup(poly_item)

    @classmethod
    def render(cls, layer: 'Layer'):
        """
        Add each symbol item group to the Scene for display

        :param layer: Draw on this layer
        """
        for s in layer.Symbols:
            # _logger.info(f"> Polygon at: [{p.vertices}]")

            # Display the item in the Qt scene
            layer.Scene.addItem(s)




""" rectangle_se.py -- Rectangle Shape Element """

# System
import logging
from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from tabletqt.layer import Layer

# Qt
from PyQt6.QtWidgets import QGraphicsPathItem, QGraphicsRectItem
from PyQt6.QtCore import QRectF, QPointF
from PyQt6.QtGui import QPainterPath

# Tablet
import tabletqt.element as element
from tabletqt.geometry_types import Rect_Size, Position
from tabletqt.styledb import StyleDB
from tabletqt.graphics.crayon_box import CrayonBox
from tabletqt.exceptions import MissingConfigData

_logger = logging.getLogger(__name__)

class RectangleSE:
    """
    Manage the rendering of Rectangle Shape Elements

    Attributes and relationships defined on the class model

    Subclass of Closed Shape on class model (R22)

    - ID {I} -- Element ID, unique within a Layer, implemented as object reference
    - Layer {I, R22} -- Element drawn on this Layer via R22/R12/R15/Element/R19/Layer
    - Size -- Rectangle height and width
    - Lower left -- Position of corner in tablet coordinates
    """
    @classmethod
    def add(cls, layer: 'Layer', asset: str, lower_left: Position, size: Rect_Size,
            color_usage: Optional['str'] = None):
        """
        Adds a rectangle to the Layer with position converted to Tablet (device) coordinates. If one is specified
        for this Shape Presentation (Presentation x Asset: See R4,R5,R21 in the class diagram), a Closed Shape Fill
        will be applied. If a color_usage is supplied, the associated color overrides the Closed Shape Fill.

        :param layer: Draw on this Layer
        :param asset: Used to determine draw style
        :param lower_left:  Lower left corner position in tablet coordinates
        :param size: The Size of the rectangle in points
        :param color_usage: If a usage string is provided, it will be indexed into the Style DB usage table to find
                            an associated RGB color.
        """
        # Flip lower left corner to device coordinates
        ll_dc = layer.Tablet.to_dc(Position(x=lower_left.x, y=lower_left.y))

        # Use upper left corner instead
        ul = Position(x=ll_dc.x, y=ll_dc.y - size.height)

        # Get the rectangle style info
        rect_style = layer.Presentation.Rectangle_presentation[asset]

        # Get the fill is defined for this rectangle asset in the layer's presentation
        # It is okay to specify no fill at all
        fill = rect_style.get('fill')

        try:
            border = rect_style['line style']
        except KeyError:
            _logger.exception(f"No line style specified in rectangle presentation for asset: [{asset}")
            raise MissingConfigData

        # If there is an overriding color usage use it instead
        if color_usage:
            try:
                fill = StyleDB.color_usage[color_usage]  # Overrides any closed shape fill
            except KeyError:
                _logger.warning(f'No color defined for usage [{color_usage}]')

        # Set the corner spec, if any
        cspec = rect_style.get('corner spec')
        # If no corner spec, assume 0 radius corners
        try:
            radius, top, bottom = (0, False, False) if not cspec else (cspec['radius'], cspec['top'], cspec['bottom'])
        except KeyError:
            _logger.exception(f"Missing data in rectangle presentation corner spec defined for asset: [{asset}")
            raise MissingConfigData

        layer.Rectangles.append(element.Rectangle(
            upper_left=ul, size=size, border_style=border, fill=fill,
            radius=radius, top=top, bottom=bottom
        ))

    @classmethod
    def roundrect(cls, x: float, y: float, width: float, height: float, top_r: int, bottom_r: int) -> QGraphicsPathItem:
        """
        Draw rectangle with rounded corners on top, bottom or both. Radius is expressed in points
        with zero resulting in a square corner on top, bottom or both

        :param x: Upper left x
        :param y: Upper left y
        :param width: Rect width
        :param height: Rect height
        :param top_r: Top corner radius
        :param bottom_r: Bottom corner radius
        """
        rect = QRectF(x, y, width, height)
        lbox = QRectF(x,y, top_r*2, bottom_r*2)
        rrect = QRectF(x + width - top_r, y, width, height)
        path = QPainterPath()

        if top_r and bottom_r:
            path.addRoundedRect(rect, bottom_r, bottom_r)

        if top_r and not bottom_r:
            path.moveTo(rect.topLeft() + QPointF(top_r, 0))
            path.lineTo(rect.right() - top_r * 2, rect.top())
            path.arcTo(rect.right() - top_r * 2, rect.top(), top_r * 2, top_r * 2, 90, -90)
            path.moveTo(rect.right(), rect.top() + top_r)
            path.lineTo(rect.bottomRight())
            path.lineTo(rect.bottomLeft())
            path.lineTo(rect.left(), rect.top() + top_r)
            path.arcTo(rect.left(), rect.top(), top_r * 2, top_r * 2, 180, -90)

        if bottom_r and not top_r:
            path.moveTo(rect.left(), rect.bottom() - bottom_r)
            path.lineTo(rect.topLeft())
            path.lineTo(rect.topRight())
            path.lineTo(rect.right(), rect.bottom() - bottom_r)
            path.arcTo(rect.right() - bottom_r * 2, rect.bottom() - bottom_r * 2, bottom_r * 2, bottom_r * 2, 0, -90)
            path.moveTo(rect.right() - bottom_r, rect.bottom())
            path.lineTo(rect.left() + bottom_r, rect.bottom())
            path.arcTo(rect.left(), rect.bottom() - bottom_r * 2, bottom_r * 2, bottom_r * 2, 270, -90)

        return QGraphicsPathItem(path)

    @classmethod
    def render(cls, layer: 'Layer'):
        """
        Draw all rectangle shapes added to a Layer

        :param layer: Draw on this Layer
        """
        for r in layer.Rectangles:
            # Diagnostic shaded rectangle bounding box under the rendered rectangle
            # to see if the corners are drawn correctly
            # frame = QGraphicsRectItem(QRectF(r.upper_left.x, r.upper_left.y, r.size.width, r.size.height))
            # fpen = QPen(Qt.GlobalColor.blue)
            # fbrush = QBrush(Qt.GlobalColor.lightGray)
            # frame.setBrush(fbrush)
            # frame.setPen(fpen)
            # layer.Scene.addItem(frame)

            # Set rectangle extents and draw
            top_radius = r.radius if r.top else 0
            bottom_radius = r.radius if r.bottom else 0
            if not top_radius and not bottom_radius:
                rect = QRectF(r.upper_left.x, r.upper_left.y, r.size.width, r.size.height)
                r_item = QGraphicsRectItem(rect)
                _logger.info(f"> Rectangle at: {r.upper_left}, size: {r.size},"
                             f"top round: {top_radius}, bottom_round: {bottom_radius}")
            else:
                r_item = cls.roundrect(r.upper_left.x, r.upper_left.y, r.size.width, r.size.height, top_radius, bottom_radius)
                _logger.info(f"> Roundrect at: {r.upper_left}, size: {r.size},"
                             f"top round: {top_radius}, bottom_round: {bottom_radius}")
            # Set pen and brush
            CrayonBox.choose_crayons(item=r_item, border_style=r.border_style, fill=r.fill)
            layer.Scene.addItem(r_item)

    @classmethod
    def render_fillrect(cls, layer: 'Layer', frect: element.FillRect):
        """
        Render a single specified filled retangle

        For now, this feature is not directly available to the user. It is used
        on demand as a component of other elements such as text to create an
        opaque underlay.

        :param layer: Draw on this layer
        :param frect: The fill color
        """
        # Create the rect item
        rect = QRectF(frect.upper_left.x, frect.upper_left.y, frect.size.width, frect.size.height)
        r_item = QGraphicsRectItem(rect)
        _logger.info(f"> Filled rect at: {frect.upper_left}, size: {frect.size}")

        # Set pen and brush
        CrayonBox.choose_fill_only(item=r_item, fill=frect.color)
        layer.Scene.addItem(r_item)



""" test_xuml_stickers.py - Draw all of the xuml class diagram stickers for different presentations """

import inspect
from pathlib import Path
from tabletqt.tablet import Tablet
from tabletqt.geometry_types import Rect_Size, Position
from tabletqt.graphics.image import ImageDE

points_in_mm = 2.83465
A4 = Rect_Size(round(210 * points_in_mm), round(297 * points_in_mm))

def test_images():
    dtype = "Starr class diagram"
    function_name = inspect.currentframe().f_code.co_name
    output_path = Path(f"output/{function_name}.pdf")

    test_tablet = Tablet(size=A4, output_file=output_path, drawing_type=dtype,
                         presentation="default", layer="diagram", show_window=False, background_color='blue steel')
    dlayer = test_tablet.layers['diagram']

    ImageDE.add(layer=dlayer, name=f"mint-small", lower_left=Position(100, 50), size=Rect_Size(180, 24))
    ImageDE.add(layer=dlayer, name=f"mint-large", lower_left=Position(100, 150), size=Rect_Size(191, 25))
    ImageDE.add(layer=dlayer, name=f"MIT-small", lower_left=Position(100, 350), size=Rect_Size(180, 24))

    test_tablet.render()
    assert True



""" demo7_images - Test display of images """

from pathlib import Path
from tabletqt.tablet import Tablet
from tabletqt.geometry_types import Rect_Size, Position
from tabletqt.graphics.image import ImageDE

points_in_mm = 2.83465

class SketchSymbols:
    tablet = None
    layer = None
    size = None

    @classmethod
    def make_a_tablet(cls):
        cls.size = Rect_Size(round(210*points_in_mm), round(297*points_in_mm))  # Ansi C portrait
        dtype = "Starr class diagram"
        # dtype = "xUML class diagram"
        # dtype = "xUML state machine diagram"
        output_path = Path(__file__).parent.parent.parent.parent / "working" / "demo7_images.pdf"
        cls.tablet = Tablet(size=cls.size, output_file=output_path, drawing_type=dtype,
                            presentation="default", layer="diagram", show_window=False, background_color='blue steel')


    @classmethod
    def sketch(cls):

        cls.make_a_tablet()
        dlayer = cls.tablet.layers['diagram']

        ImageDE.add(layer=dlayer, name=f"mint-small", lower_left=Position(100,50), size=Rect_Size(180,24))
        ImageDE.add(layer=dlayer, name=f"mint-large", lower_left=Position(100,150), size=Rect_Size(191,25))
        ImageDE.add(layer=dlayer, name=f"MIT-small", lower_left=Position(100,350), size=Rect_Size(180,24))
        cls.tablet.render()


if __name__ == "__main__":
    SketchSymbols.sketch()


""" scene_view.py - create a vector drawing view in QT """

# System
import logging
from PyQt6.QtWidgets import QGraphicsView, QVBoxLayout, QGraphicsScene, QWidget
from PyQt6.QtCore import Qt
from PyQt6.QtGui import QColor, QPainter, QPageSize, QPageLayout
from PyQt6.QtPrintSupport import QPrinter
from pypdf import PdfWriter, PdfReader

_logger = logging.getLogger(__name__)

# A generated PDF leave extra space on the bottom or right side depending
# on the tablet dimensions. These are the crop values that seem to work
# determined experiementally for the height to width ratio for each
# standard sheet size and orientation.
# Thus, all keys less than zero are for landscape orientation
# Each corresponding value is the amount to crop the pdf on the bottom
# for landscape and off the right side for portrait
pdf_crop = { 0.65: 196, 0.71: 101, 0.75: 32, 0.77: 0, 1.29: 635, 1.33: 665, 1.55: 789}

class SceneView(QGraphicsView):
    def __init__(self, size, background):
        super().__init__()

        # Create a QGraphicsScene
        self.tablet_size = size
        self.scene = QGraphicsScene()
        self.setSceneRect(0, 0, size.width, size.height)
        pad = 10
        self.setFixedSize(size.width + pad, size.height + pad)
        if background:
            _logger.info(f"Setting scene background to RGB: {background}")
            self.scene.setBackgroundBrush(QColor(*background))

        # Set the scene to the view
        self.setScene(self.scene)

        self.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)

        self.setTransformationAnchor(QGraphicsView.ViewportAnchor.NoAnchor)
        self.setResizeAnchor(QGraphicsView.ViewportAnchor.NoAnchor)
        # self.setRenderHint(QPainter.RenderHint.Antialiasing)
        # self.setRenderHint(QPainter.RenderHint.TextAntialiasing)

    def save_as_pdf(self, file_path):
        # Create a QPrinter object and set the output format to PDF
        printer = QPrinter(QPrinter.PrinterMode.PrinterResolution)
        printer.setOutputFormat(QPrinter.OutputFormat.PdfFormat)
        # printer.setPageSize(QPageSize(self.size))
        ps = QPageSize(QPageSize.PageSizeId.AnsiC)
        printer.setPageSize(ps)
        printer.setPageOrientation(QPageLayout.Orientation.Landscape)
        # printer.setPageMargins(QMarginsF(0, 0, 0, 0))
        printer.setFullPage(True)
        printer.setOutputFileName(str(file_path))

        # Create a QPainter to render the view content onto the printer
        painter = QPainter(printer)

        # Render the entire scene to the printer (the PDF)
        self.render(painter)

        # Finish the painting process
        painter.end()

        self.crop_pdf(file_path)

    def crop_pdf(self, file_path):
        reader = PdfReader(file_path)
        writer = PdfWriter()
        hw_ratio = self.tablet_size.height / self.tablet_size.width

        sheet = reader.pages[0]
        crop_box = sheet.cropbox

        # Find the closest key by selecting the key with minimum distance
        # to the hw_ratio
        closest_key = min(pdf_crop.keys(), key=lambda k: abs(k - hw_ratio))

        # And that gives us an experimentally derived crop value
        crop = pdf_crop[closest_key]

        # I did compute a linear function based on the dictionary values that comes close
        # to the crop data set, but is sometimes off by a little too much.
        # So I left it commented out here for reference, but use the data set directly to get the best crop value
        if self.tablet_size.height > self.tablet_size.width: # portrait
            # crop = 582.65 * hw_ratio - 113.55
            crop_box.upper_right = (crop_box.upper_right[0] - crop, crop_box.upper_right[1])
            sheet.cropbox.upper_right = crop_box.upper_right
        else: # landscape
            # crop = -1635.53 * hw_ratio + 1259.99
            crop_box.lower_left = (crop_box.lower_left[0], crop_box.lower_left[1] + crop)
            sheet.cropbox.lower_left = crop_box.lower_left

        writer.add_page(sheet)
        with open(file_path, "wb") as fp:
            writer.write(fp)
        # with open("cropped.pdf", "wb") as fp:
        #     writer.write(fp)


class MainWindow(QWidget):
    def __init__(self, title, size, background):
        super().__init__()

        # Set up the layout and graphics view
        layout = QVBoxLayout()
        self.graphics_view = SceneView(size, background)
        layout.addWidget(self.graphics_view)

        self.setLayout(layout)
        self.setWindowTitle(title)
        self.setGeometry(500, 100, size.width, size.height)  # Position and size of the main window on the screen



""" demo8_grid_rect.py - Test grid diagnostic boundary rect """

from pathlib import Path
from tabletqt.tablet import Tablet
from tabletqt.geometry_types import Rect_Size, Position
from tabletqt.graphics.symbol import Symbol
from tabletqt.graphics.text_element import TextElement, TextBlockCorner, HorizAlign
from tabletqt.graphics.rectangle_se import RectangleSE
from tabletqt.graphics.line_segment import LineSegment
from tabletqt.graphics.diagnostic_marker import DiagnosticMarker
from tabletqt.exceptions import TabletBoundsExceeded

points_in_mm = 2.83465

class SketchSymbols:
    tablet = None
    layer = None
    size = None

    @classmethod
    def make_a_tablet(cls):
        cls.size = Rect_Size(height=round(841*points_in_mm), width=round(1189*points_in_mm))  # Ansi C portrait
        dtype = "Grid Diagnostic"
        output_path = Path(__file__).parent.parent.parent.parent / "working" / "demo8_grid_rect.pdf"
        cls.tablet = Tablet(size=cls.size, output_file=output_path, drawing_type=dtype,
                            presentation="default", layer="grid", show_window=False, background_color='white')

    @classmethod
    def sketch(cls):

        cls.make_a_tablet()
        glayer = cls.tablet.layers['grid']

        try:
            LineSegment.add(layer=glayer, asset="row boundary", from_here=Position(210, 2398.64),
                            to_there=Position(2772, 2398.6499))
        except TabletBoundsExceeded as e:
            pass

        # r = RectangleSE.add(layer=glayer, asset="grid boundary", lower_left=Position(2398.64, 210),
        #                     size=Rect_Size(height=2398.64, width=210))
        cls.tablet.render()

if __name__ == "__main__":
    SketchSymbols.sketch()



""" line_segment.py -- Line Segment """

# System
import logging
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from tabletqt.layer import Layer

# Qt
from PyQt6.QtWidgets import QGraphicsLineItem
from PyQt6.QtCore import QLineF

# Tablet
import tabletqt.element as element
from tabletqt.geometry_types import Position
from tabletqt.graphics.crayon_box import CrayonBox
from tabletqt.exceptions import MissingConfigData

_logger = logging.getLogger(__name__)
class LineSegment:
    """
    Manage the rendering of Line Segments

    Attributes and relationships defined on the class model

    Subclass of Shape Element on class model (R12)

    - ID {I} -- Element ID, unique within a Layer, implemented as object reference
    - Layer {I, R12} -- Element drawn on this Layer via R12/R15/Element/R19/Layer
    - From -- Draw from here in tablet coordinates
    - To -- To here in tablet coordinates
    """
    @classmethod
    def add(cls, layer: 'Layer', asset: str, from_here: Position, to_there: Position):
        """
        Convert line segment coordinates to device coordinates and combine with the Line Style defined
        for the Asset in the selected Preentation Style

        :param layer: Draw on this layer
        :param asset: Used to determine draw style
        :param from_here: In tablet coordinates
        :param to_there: In tablet coordinates
        """
        try:
            line_style = layer.Presentation.Line_presentation[asset]['line style']
        except KeyError:
            _logger.exception(f"No line style specified in line presentation for asset: [{asset}")
            raise MissingConfigData

        layer.Line_segments.append(
            element.Line_Segment(from_here=layer.Tablet.to_dc(from_here), to_there=layer.Tablet.to_dc(to_there),
                                 style=line_style)
        )

    @classmethod
    def render(cls, layer: 'Layer'):
        """
        Draw the line segments

        :param layer: Draw on this layer
        """
        for ls in layer.Line_segments:
            _logger.info(f"> Line {ls.from_here}, {ls.to_there}")

            # Create a line item for the scene
            ls_item = QGraphicsLineItem(QLineF(*ls.from_here, *ls.to_there))
            CrayonBox.choose_crayons(item=ls_item, border_style=ls.style)
            layer.Scene.addItem(ls_item)



""" styles.py - Named tuples representing various style data """

from typing import NamedTuple

class FloatRGB(NamedTuple):
    r: int
    g: int
    b: int

class LineStyle(NamedTuple):
    pattern: str
    width: int
    color: str

class TextStyle(NamedTuple):
    typeface: str
    size: int
    slant: str
    weight: str
    color: str
    spacing: float

class DashPattern(NamedTuple):
    solid: float
    blank: float



""" etchasketch.py - Testbed for exercising the tabletqt with drawing patterns """

from pathlib import Path
from tabletqt.tablet import Tablet
from tabletqt.geometry_types import Rect_Size, Position
from tabletqt.graphics.circle_se import CircleSE
from tabletqt.graphics.polygon_se import PolygonSE
from tabletqt.graphics.line_segment import LineSegment
from tabletqt.graphics.rectangle_se import RectangleSE
from tabletqt.graphics.image import ImageDE
from tabletqt.graphics.text_element import TextElement
from tabletqt.graphics.diagnostic_marker import DiagnosticMarker

points_in_mm = 2.83465
points_in_cm = 28.3465
points_in_inch = 72

class EtchaSketch:
    tablet = None
    layer = None
    size = None

    @classmethod
    def make_a_tablet(cls):
        # cls.size = Rect_Size(int(297*points_in_mm), int(420*points_in_mm)) # A3
        # cls.size = Rect_Size(int(594*points_in_mm), int(841*points_in_mm)) # A1
        # cls.size = Rect_Size(int(420*points_in_mm), int(594*points_in_mm)) # A2
        # cls.size = Rect_Size(int(8.5*72), 11*72) # letter landscape
        # cls.size = Rect_Size(11*72, int(8.5*72)) # letter portrait
        # cls.size = Rect_Size(11*72, 17*72) # Ansi B landscape
        # cls.size = Rect_Size(17*72, 11*72) # Ansi B portrait
        # cls.size = Rect_Size(22*72, 17*72)  # Ansi C portrait
        # cls.size = Rect_Size(34*72, 22*72) # Ansi D portrait
        cls.size = Rect_Size(22*72, 34*72) # Ansi D
        # cls.size = Rect_Size(48*72, 36*72) # Ansi E portrait
        # cls.size = Rect_Size(36*72, 48*72) # Ansi E
        output_path = Path(__file__).parent.parent.parent / "working" / "tabletqt.pdf"
        cls.tablet = Tablet(app='tablet', size=cls.size, output_file=output_path, drawing_type="xUML State Machine Diagram",
                            presentation="default", layer="diagram", background_color='white')

    @classmethod
    def draw_tabloid(cls):
        cls.make_a_tablet()
        dlayer = cls.tablet.layers['diagram']

        RectangleSE.add(layer=dlayer, asset='state name compartment',
                        lower_left=Position(50, 300), size=Rect_Size(height=20, width=70))

        PolygonSE.add(layer=dlayer, asset='solid arrow', vertices=[
            Position(300, 300),
            Position(320, 340),
            Position(340, 300)
        ])
        pad = 0
        # Horizontal top
        LineSegment.add(layer=dlayer, asset='transition connector',
                        from_here=Position(pad, cls.size.height),
                        to_there=Position(cls.size.width, cls.size.height))

        # Horizontal bottom
        LineSegment.add(layer=dlayer, asset='transition connector',
                        from_here=Position(pad, pad),
                        to_there=Position(cls.size.width, pad))
        # vertical left
        LineSegment.add(layer=dlayer, asset='transition connector',
                        from_here=Position(pad, cls.size.height),
                        to_there=Position(pad, pad))

        # vertical right
        LineSegment.add(layer=dlayer, asset='transition connector',
                        from_here=Position(cls.size.width, cls.size.height),
                        to_there=Position(cls.size.width, pad))

        cls.tablet.render()

    @classmethod
    def draw_stuff(cls):
        cls.make_a_tablet()

        dlayer = cls.tablet.layers['diagram']

        slayer = cls.tablet.add_layer(name='sheet', presentation='default',
                                      drawing_type="Model Integration Diagram Large Frame")

        # TextElement.add_block(layer=slayer, asset='title', lower_left=Position(500, 500),
        #                       text=['From here', 'to quick goats'])

        RectangleSE.add(layer=slayer, asset='block border',
                        lower_left=Position(1150, 150), size=Rect_Size(100, 300))

        RectangleSE.add(layer=dlayer, asset='state name compartment',
                        lower_left=Position(50, 700), size=Rect_Size(height=20, width=70))

        PolygonSE.add(layer=dlayer, asset='solid arrow', vertices=[
            Position(300, 300),
            Position(320, 340),
            Position(340, 300)
        ])

        ImageDE.add(layer=slayer, name='mint-medium', lower_left=Position(1150, 50), size=Rect_Size(10, 10))

        RectangleSE.add(layer=dlayer, asset='state name only compartment', lower_left=Position(200, 900),
                                                   size=Rect_Size(30, 150))

        pad = 0
        # Horizontal top
        LineSegment.add(layer=dlayer, asset='transition connector',
                        from_here=Position(pad, cls.size.height),
                        to_there=Position(cls.size.width, cls.size.height))

        # Horizontal bottom
        LineSegment.add(layer=dlayer, asset='transition connector',
                        from_here=Position(pad, pad),
                        to_there=Position(cls.size.width, pad))
        # vertical left
        LineSegment.add(layer=dlayer, asset='transition connector',
                        from_here=Position(pad, cls.size.height),
                        to_there=Position(pad, pad))

        # vertical right
        LineSegment.add(layer=dlayer, asset='transition connector',
                        from_here=Position(cls.size.width, cls.size.height),
                        to_there=Position(cls.size.width, pad))

        LineSegment.add(layer=dlayer, asset='transition connector',
                        from_here=Position(510, 400), to_there=Position(510, 600))

        LineSegment.add(layer=dlayer, asset='transition connector',
                        from_here=Position(230, 580), to_there=Position(230, 200))

        CircleSE.add(layer=dlayer, asset='solid small dot', center=Position(400, 400), radius=20)

        # General pin reference
        DiagnosticMarker.add_cross_hair(dlayer, Position(200, 500))
        DiagnosticMarker.add_cross_hair(dlayer, Position(200, 600))
        DiagnosticMarker.add_cross_hair(dlayer, Position(200, 700))

        # Print this text
        # sample_text = "Thirty seven goodies"
        # cls.tabletqt.layers['diagram'].add_text_line(
        #     asset='transition name', lower_left=Position(255, 623), text=sample_text)
        sample_text = "Thirty seven goodies"
        TextElement.add_line(layer=dlayer,
                             asset='state name', lower_left=Position(215, 300), text=sample_text)

        DiagnosticMarker.add_cross_hair(dlayer, Position(498, 712), 'purple')
        # Get the tbox size
        # tbox = cls.tabletqt.layers['diagram'].text_line_size(asset='transition name', text_line=sample_text)
        tbox = TextElement.line_size(layer=dlayer, asset='state name', text_line=sample_text)

        # Draw cross hairs where the displayed text should line up
        DiagnosticMarker.add_cross_hair(dlayer, Position(255, 623), 'green')  # Lower left of text
        DiagnosticMarker.add_cross_hair(dlayer, Position(255, 623 + tbox.height), 'green')  # Upper left of text

        # Put bounding box at the upper left crosshair
        DiagnosticMarker.add_raw_rectangle(layer=dlayer, upper_left=Position(255, 623 + tbox.height),
                                           size=Rect_Size(tbox.height, tbox.width))

        pad_x = 5  # To adjust the position of the bounding rectangle relative to the text upper left corner
        pad_y = -4
        # cls.tabletqt.layers['diagram'].add_raw_rectangle(upper_left=Position(255, 635+pad_y),
        #                                                size=Rect_Size(tbox.height, tbox.width))
        # DiagnosticMarker.add_cross_hair(layer=dlayer, Position(255, 635)) # Upper left of text
        #
        # DiagnosticMarker.add_cross_hair(layer=dlayer, Position(5, 5))
        # DiagnosticMarker.add_cross_hair(layer=dlayer, Position(255, 623), 'green') # Lower left of text
        # DiagnosticMarker.add_cross_hair(layer=dlayer, Position(255, 623+tbox.height), 'red')  # Upper left of text

        # Lower left of text block
        DiagnosticMarker.add_cross_hair(dlayer, Position(500, 500), 'green')
        TextElement.add_block(layer=dlayer, asset='transition name', lower_left=Position(500, 500),
                              text=['From here', 'to quick goats'])
        tbox = TextElement.text_block_size(layer=dlayer, asset='transition name',
                                           text_block=['From here', 'to quick goats'])
        # cls.tabletqt.layers['diagram'].add_raw_rectangle(upper_left=Position(500, 500+tbox.height),
        #                                                size=Rect_Size(tbox.height, tbox.width))

        cls.tablet.render()



""" test_starrcd_symbols.py - Draw all of the Starr class diagram symbols for different presentation """

import pytest
import inspect
from pathlib import Path
from tabletqt.tablet import Tablet
from tabletqt.geometry_types import Rect_Size, Position
from tabletqt.graphics.symbol import Symbol

points_in_mm = 2.83465
A4 = Rect_Size(round(210 * points_in_mm), round(297 * points_in_mm))

def test_starrcd_symbols():
    dtype = "Starr class diagram"
    function_name = inspect.currentframe().f_code.co_name
    output_path = Path(f"output/{function_name}.pdf")

    test_tablet = Tablet(size=A4, output_file=output_path, drawing_type=dtype,
                         presentation="default", layer="diagram", show_window=False, background_color='autumn')
    dlayer = test_tablet.layers['diagram']

    s = Symbol(layer=dlayer, name='ordinal', pin=Position(50, 400), angle=0)
    s = Symbol(layer=dlayer, name='1 mult', pin=Position(75, 400), angle=0)
    s = Symbol(layer=dlayer, name='M mult', pin=Position(100, 400), angle=0)
    s = Symbol(layer=dlayer, name='1c mult', pin=Position(125, 400), angle=0)
    s = Symbol(layer=dlayer, name='Mc mult', pin=Position(150, 400), angle=0)
    s = Symbol(layer=dlayer, name='superclass', pin=Position(175, 400), angle=0)

    s = Symbol(layer=dlayer, name='ordinal', pin=Position(50, 300), angle=90)
    s = Symbol(layer=dlayer, name='1 mult', pin=Position(75, 300), angle=90)
    s = Symbol(layer=dlayer, name='M mult', pin=Position(100, 300), angle=90)
    s = Symbol(layer=dlayer, name='1c mult', pin=Position(125, 300), angle=90)
    s = Symbol(layer=dlayer, name='Mc mult', pin=Position(150, 300), angle=90)
    s = Symbol(layer=dlayer, name='superclass', pin=Position(175, 300), angle=90)

    test_tablet.render()
    assert True

def test_bp_starrcd_symbols():
    dtype = "Starr class diagram"
    function_name = inspect.currentframe().f_code.co_name
    output_path = Path(f"output/{function_name}.pdf")

    test_tablet = Tablet(size=A4, output_file=output_path, drawing_type=dtype,
                         presentation="blueprint", layer="diagram", show_window=False, background_color='blueprint')
    dlayer = test_tablet.layers['diagram']

    s = Symbol(layer=dlayer, name='ordinal', pin=Position(50, 400), angle=0)
    s = Symbol(layer=dlayer, name='1 mult', pin=Position(75, 400), angle=0)
    s = Symbol(layer=dlayer, name='M mult', pin=Position(100, 400), angle=0)
    s = Symbol(layer=dlayer, name='1c mult', pin=Position(125, 400), angle=0)
    s = Symbol(layer=dlayer, name='Mc mult', pin=Position(150, 400), angle=0)
    s = Symbol(layer=dlayer, name='superclass', pin=Position(175, 400), angle=0)

    s = Symbol(layer=dlayer, name='ordinal', pin=Position(50, 300), angle=90)
    s = Symbol(layer=dlayer, name='1 mult', pin=Position(75, 300), angle=90)
    s = Symbol(layer=dlayer, name='M mult', pin=Position(100, 300), angle=90)
    s = Symbol(layer=dlayer, name='1c mult', pin=Position(125, 300), angle=90)
    s = Symbol(layer=dlayer, name='Mc mult', pin=Position(150, 300), angle=90)
    s = Symbol(layer=dlayer, name='superclass', pin=Position(175, 300), angle=90)

    test_tablet.render()
    assert True




""" text_element.py -- Text Element """

# System
import logging
from pathlib import Path
from enum import Enum
from typing import TYPE_CHECKING, List

if TYPE_CHECKING:
    from tabletqt.layer import Layer
    from tabletqt.presentation import Presentation

# Qt
from PyQt6.QtWidgets import QGraphicsTextItem
from PyQt6.QtGui import QFont, QFontMetrics, QColor

# Modelint
from mi_config.config import Config

# Tablet
import tabletqt.element as element
from tabletqt.geometry_types import Position, Rect_Size, HorizAlign
from tabletqt.styledb import StyleDB
from tabletqt.graphics.rectangle_se import RectangleSE
from tabletqt.exceptions import TabletBoundsExceeded
from tabletqt.tablet_config import TabletConfig

logger = logging.getLogger(__name__)

# Constants
tbox_xoffset = 4  # QT distance from text item origin (setPos) to lower left corner of text bounding box
tbox_yoffset = 4  # For now, determined experimentally, not sure how to compute it from QFontMetrics

underlay_margin_h = 5  # Horizontal and vertical distances from text to outer edge of underlay
underlay_margin_v = 5

underlay_offset_x = 2  # Lower left corner offsets from lower left origin of text item origin
underlay_offset_y = 3

Qt_font_weight = {'normal': QFont.Weight.Normal, 'bold': QFont.Weight.Bold}
"""Maps an application style to a Qt specific font weight"""


class TextBlockCorner(Enum):
    """ Text Block corners """
    UL = "upper left"
    UR = "upper right"
    LL = "lower left"
    LR = "lower right"


class TextElement:
    """
    Manage the defnition and rendering of Text Elements (blocks of text)

    Attributes and relationships defined on the class model

    Subclass of Element on class model (R15)

    - ID {I} -- Element ID, unique within a Layer, implemented as object reference
    - Layer {I, R22} -- Element drawn on this Layer via R22/R12/R15/Element/R19/Layer
    - Content -- One or more lines of text, implemented here as a list of strings
    - Lower left -- Position in tablet coordinates of the entire text block
    - Text style {R16} -- Typeface, color, and other display properties to be applied
    """

    stickers = None

    @classmethod
    def load_stickers(cls):
        """
        Load all predefined sticker text from the stickers.yaml file
        """
        sticker_data = Config(app_name=TabletConfig.app_name, lib_config_dir=TabletConfig.config_path,
                              fspec={'stickers': None})
        cls.stickers = sticker_data.loaded_data['stickers']

    @classmethod
    def lower_left_pin(cls, presentation: 'Presentation', asset: str, text_block: List[str], pin: Position,
                       corner: TextBlockCorner) -> Position:
        """
        Given a text block and the position of one of its corners, return the position of its lower left corner

        :param presentation: The target layer's Presentation
        :param asset:  Name of asset that defines the style properties for rendering the text
        :param text_block:  Name of the sticker to be applied
        :param pin: Location of some corner (like a thumbtack pin) in Tablet coordinates
        :param corner: The corner to be pinned (upper left, lower right, ...)
        :return: The position of the lower left corner, also in Tablet coordinates
        """
        # Check the simple case first
        if corner == TextBlockCorner.LL:
            return pin  # No need to compute, we already have a lower left corner

        # First we need the size of the text block with asset specified styling applied
        tblock_size = cls.text_block_size(presentation=presentation, asset=asset, text_block=text_block)

        # Compute from the other three possible corners
        if corner == TextBlockCorner.UL:  # upper left
            return Position(pin.x, pin.y - tblock_size.height)
        if corner == TextBlockCorner.LR:  # lower right
            return Position(pin.x - tblock_size.width, pin.y)
        if corner == TextBlockCorner.UR:  # upper right
            return Position(pin.x - tblock_size.width, pin.y - tblock_size.height)

    @classmethod
    def add_sticker(cls, layer: 'Layer', asset: str, name: str, pin: Position, corner: TextBlockCorner):
        """
        Place a Sticker at the requested location. The client app will provide a corner and a position, along with
        the name of the Sticker, its style properties (asset) and the Layer where the Sticker will be added.

        A sticker is always a singe line of text, so we don't need alignment specified

        :param layer:  Place sticker on this layer
        :param asset:  Defines the style properties for rendering the text
        :param name:  Name of the sticker to be applied
        :param pin: Location of some corner (like a thumbtack pin)
        :param corner: The corner to be pinned (upper left, lower right, ...)
        :return:
        """

        drawing_type = cls.stickers[layer.Drawing_type]
        try:
            sticker_text = drawing_type[asset][name]
        except KeyError:
            # No sticker defined for this asset and name
            return

        # Resolve position
        ll_corner = cls.lower_left_pin(presentation=layer.Presentation, asset=asset, text_block=[sticker_text],
                                       pin=pin, corner=corner)

        # Add the sticker content as a line of text (sticker's are always a single line of text)
        cls.add_line(layer=layer, asset=asset, lower_left=ll_corner, text=sticker_text)

    @classmethod
    def line_size(cls, presentation: 'Presentation', asset: str, text_line: str) -> Rect_Size:
        """
        Returns the size of a line of text when rendered with the asset's text style

        :param presentation:  The Presentation
        :param asset: Determines text display style
        :param text_line: Line of text
        :return: Size of the text line ink area
        """
        style_name = presentation.Text_presentation[asset]  # Look up the text style for this asset
        style = StyleDB.text_style[style_name['text style']]
        font_name = StyleDB.typeface[style.typeface]

        font = QFont(font_name, style.size)
        font.setItalic(style.slant == 'italic')
        font.setBold(style.weight == 'bold')

        font_metrics = QFontMetrics(font)
        bound_rect = font_metrics.boundingRect(text_line)

        return Rect_Size(height=bound_rect.height(), width=bound_rect.width())

    @classmethod
    def text_block_size(cls, presentation: 'Presentation', asset: str, text_block: List[str]) -> Rect_Size:
        """
        Determines the dimensions of a rectangle bounding the text to be drawn.

        :param presentation:
        :param asset: Name of the text asset to get display style properties
        :param text_block: A list of text lines to be displayed
        :return:  The display size of the text block
        """
        style_name = presentation.Text_presentation[asset]  # Look up the text style for this asset
        style = StyleDB.text_style[style_name['text style']]
        font_height = style.size
        spacing = font_height * style.spacing
        inter_line_spacing = spacing - font_height  # Space between two lines

        num_lines = len(text_block)
        assert num_lines > 0, "Text block size requested for empty text block"
        # The text block is the width of its widest ink render extent
        widths = [cls.line_size(presentation=presentation, asset=asset, text_line=line).width for line in text_block]
        widest_line = text_block[widths.index(max(widths))]
        block_width = cls.line_size(presentation=presentation, asset=asset, text_line=widest_line).width
        block_height = num_lines * spacing - inter_line_spacing  # Deduct that one unneeded line of spacing on the top

        return Rect_Size(width=block_width, height=block_height)

    @classmethod
    def add_underlay(cls, layer: 'Layer', lower_left: Position, size: Rect_Size):
        """
        Adds a rectangle matching the tablet background color. This rectangle will be
        drawn underneath a text line or block so that the color surrounding
        the text matches the background. This is useful when you want to draw text
        over the top of some graphical component such as a line without being too
        visually disruptive.

        :param layer: Draw on this layer
        :param lower_left: Lower left corner position in tablet coordinates
        :param size: The Size of the rectangle in points
        """
        # Flip lower left corner to device coordinates
        ll_dc = layer.Tablet.to_dc(Position(x=lower_left.x, y=lower_left.y))

        # Use upper left corner instead
        ul = Position(x=ll_dc.x, y=ll_dc.y - size.height)

        layer.TextUnderlayRects.append(element.FillRect(
            upper_left=ul, size=size, color=layer.Tablet.background_color))

    @classmethod
    def add_line(cls, layer: 'Layer', asset: str, lower_left: Position, text: str):
        """
        Adds a line of text to the Tablet at the specified lower left corner pin which will be converted
        to device coordinates

        :param layer: Draw on this layer
        :param asset: Used to determine text style
        :param lower_left: Lower left corner position of text line in tablet coordinates
        :param size: The size of the text line rectangle in points
        """
        # Qt positions using the upper left corner
        # We need to determine the height of the text bounding box to determine the upper left corner
        ll_dc = layer.Tablet.to_dc(Position(x=lower_left.x, y=lower_left.y))  # Convert to device coordinates
        tl_size = cls.line_size(presentation=layer.Presentation, asset=asset,
                                text_line=text)  # Get height of bounding box
        # Compute upper left corner using experimentally observed offset
        ul = Position(x=ll_dc.x - tbox_xoffset, y=ll_dc.y - tl_size.height - tbox_yoffset)

        if asset in layer.Presentation.Underlays:
            # Compute a rectangle slightly larger than the text area to underlay the text
            underlay_size = Rect_Size(height=tl_size.height + underlay_margin_v,
                                      width=tl_size.width + underlay_margin_h)
            underlay_pos = Position(lower_left.x - underlay_offset_x, lower_left.y - underlay_offset_y)
            cls.add_underlay(layer=layer, lower_left=underlay_pos, size=underlay_size)
        try:
            layer.Text.append(
                element.Text_line(
                    upper_left=ul, text=text,
                    style=layer.Presentation.Text_presentation[asset],
                )
            )
        except TabletBoundsExceeded:
            logger.warning(f"Asset: [{asset}] Text: [{text}] outside of tabletlib draw area")
            return

    @classmethod
    def pin_block(cls, layer: 'Layer', asset: str, text: List[str], pin: Position,
                  corner: TextBlockCorner, align: HorizAlign = HorizAlign.LEFT):
        """
        Pin a text block to a Layer specifying any of the four block corners. This simplifies
        box placement computations for the user.

        :param layer:
        :param asset:
        :param lower_left:
        :param text:
        :param pin:
        :param corner:
        :param align:
        """
        # Resolve position
        ll_corner = cls.lower_left_pin(presentation=layer.Presentation, asset=asset, text_block=text,
                                       pin=pin, corner=corner)
        cls.add_block(layer=layer, asset=asset, lower_left=ll_corner, text=text, align=align)

    @classmethod
    def add_block(cls, layer: 'Layer', asset: str, lower_left: Position, text: List[str],
                  align: HorizAlign = HorizAlign.LEFT):
        """
        Add all lines in this block of text to the Tablet.

        Set the lower left x of each line based on right or left alignment
        within the text block.  Assuming left alignment as default.

        :param layer:  Add text to this Layer
        :param asset:  To get the text style
        :param lower_left: Lower left corner of the text block on the Tablet
        :param text: One or more lines of text
        :param align: Horizontal text alignment (left, right or center)
        """
        style_name = layer.Presentation.Text_presentation[asset]  # Look up the text style for this asset
        style = StyleDB.text_style[style_name['text style']]
        font_height = style.size
        spacing = font_height * style.spacing

        # Get height of one line (any will do since they all use the same text style)
        xpos, ypos = lower_left  # Initialize at lower left corner
        x_indent = 0  # Assumption for left aligned block
        block_width = None
        if align != HorizAlign.LEFT:
            # We'll need the total width of the block as a reference point
            longest_line = max(text, key=len)
            block_width = cls.line_size(presentation=layer.Presentation, asset=asset, text_line=longest_line).width
        for line in text[::-1]:  # Reverse order since we are positioning lines from the bottom up
            # always zero indent from xpos when left aligned
            if align == HorizAlign.RIGHT:
                assert block_width, "block_width not set"
                line_width = cls.line_size(presentation=layer.Presentation, asset=asset, text_line=line).width
                x_indent = block_width - line_width  # indent past xpos by the difference
            if align == HorizAlign.CENTER:
                line_width = cls.line_size(presentation=layer.Presentation, asset=asset, text_line=line).width
                x_indent = (block_width - line_width) / 2  # indent 1/2 of non text span
            cls.add_line(layer=layer, asset=asset, lower_left=Position(xpos + x_indent, ypos), text=line)
            ypos += spacing

    @classmethod
    def render_underlays(cls, layer: 'Layer'):
        """
        Draw all text underlays as filled, borderless rectangles

        :param layer: Draw on this layer
        """
        for u in layer.TextUnderlayRects:
            RectangleSE.render_fillrect(layer=layer, frect=u)

    @classmethod
    def render(cls, layer: 'Layer'):
        """
        Draw all lines of text on this layer

        :param layer: Draw on this Layer
        """
        for t in layer.Text:
            t_item = QGraphicsTextItem(t.text)
            style = StyleDB.text_style[t.style['text style']]
            text_color_name = StyleDB.text_style[t.style['text style']].color
            text_rgb_color_value = StyleDB.color[text_color_name]
            t_item.setDefaultTextColor(QColor(*text_rgb_color_value))
            font = QFont(StyleDB.typeface[style.typeface], style.size)
            font.setWeight(Qt_font_weight[style.weight])
            if style.slant == 'italic':
                font.setItalic(True)
            t_item.setFont(font)
            logger.info(f'Font [{style}]')
            t_item.setPos(t.upper_left.x, t.upper_left.y)
            logger.info(f'> Text line [{t.text}] at {t.upper_left}')
            layer.Scene.addItem(t_item)



""" demo3_shapes.py - Test shape display """

from pathlib import Path
from tabletqt.tablet import Tablet
from tabletqt.geometry_types import Rect_Size, Position
from tabletqt.graphics.symbol import Symbol
from tabletqt.graphics.text_element import TextElement, TextBlockCorner, HorizAlign
from tabletqt.graphics.rectangle_se import RectangleSE
from tabletqt.graphics.diagnostic_marker import DiagnosticMarker

points_in_mm = 2.83465
points_in_cm = 28.3465
points_in_inch = 72


class SketchSymbols:
    tablet = None
    layer = None
    size = None

    @classmethod
    def make_a_tablet(cls):
        cls.size = Rect_Size(round(210*points_in_mm), round(297*points_in_mm))  # Ansi C portrait
        dtype = "Starr class diagram"
        # dtype = "xUML class diagram"
        # dtype = "xUML state machine diagram"
        output_path = Path(__file__).parent.parent.parent.parent / "working" / "demo3_shapes.pdf"
        cls.tablet = Tablet(size=cls.size, output_file=output_path, drawing_type=dtype,
                            presentation="default", layer="diagram", show_window=False, background_color='blue steel')


    @classmethod
    def sketch(cls):

        cls.make_a_tablet()
        dlayer = cls.tablet.layers['diagram']

        vphrase = ["draws", "boundary with"]



        # TextElement.pin_block(layer=dlayer, asset='class face name', pin=Position(300,300), text=vphrase,
        #                       corner=TextBlockCorner.UR, align=HorizAlign.RIGHT)

        r = RectangleSE.add(layer=dlayer, asset="class name compartment", lower_left=Position(100, 100),
                            size=Rect_Size(height=27, width=253))
        # s = Symbol(layer=dlayer, name='final pseudo state', pin=Position(300, 300), angle=0)
        # s = Symbol(app='flatland', layer=dlayer, group='Starr class', name='M mult',
        #            pin=Position(400, 300), angle=0)

        DiagnosticMarker.add_raw_rectangle(layer=dlayer, upper_left=Position(10, 580),
                                           size=Rect_Size(width=800,height=580))
        DiagnosticMarker.add_cross_hair(layer=dlayer, location=Position(10, 580))
        # DiagnosticMarker.add_cross_hair(dlayer, Position(300, 275))
        cls.tablet.render()


if __name__ == "__main__":
    SketchSymbols.sketch()


""" polygon_se.py -- Polygon Shape Element """

# System
import logging
from typing import TYPE_CHECKING, List

# Qt
from PyQt6.QtWidgets import QGraphicsPolygonItem
from PyQt6.QtCore import QPointF
from PyQt6.QtGui import QPolygonF

# Tablet
from tabletqt.geometry_types import Position
from tabletqt.graphics.crayon_box import CrayonBox
from tabletqt.graphics.line_segment import LineSegment

if TYPE_CHECKING:
    from tabletqt.layer import Layer

_logger = logging.getLogger(__name__)
class PolygonSE:
    """
    Manage the rendering of Polygon Shape Elements which are Closed Shapes

    Additionally, a contiguous sequence of Line Segements are managed here as an
    open polygon shape. There is no concept of an open polygon in the class models so
    the add_open method could just as easily be part of the Line Segment Element class.
    In fact, there is no enforced constraint that the segments be contiguous.

    (The idea was to support drawing a pattern like an open arrow head).

    For legacy reasons, let's just leave it here for now.

    Attributes and relationships defined on the class model

    Subclass of Closed Shape on class model (R22)
    Related via {R17} to an ordered set of Vertex instances
    These are implemented as a supplied list of tablet positions

    - ID {I} -- Element ID, unique within a Layer, implemented as object reference
    - Layer {I, R22} -- Element drawn on this Layer via R22/R12/R15/Element/R19/Layer
    """

    @classmethod
    def add(cls, layer: 'Layer', asset: str, vertices: List[Position]):
        """
        Add a closed polygon to the layer

        :param layer: Draw on this layer
        :param asset: Used to determine draw style
        :param vertices: Polygon vertices in Tablet coordinates
        """
        # Flip each position to device coordinates
        device_vertices = [layer.Tablet.to_dc(v) for v in vertices]
        pverts = [QPointF(*v) for v in device_vertices]
        polygon = QPolygonF(pverts)
        poly_item = QGraphicsPolygonItem(polygon)

        # Set pen and brush
        border_style = layer.Presentation.Shape_presentation[asset]
        fill = layer.Presentation.Closed_shape_fill[asset]
        CrayonBox.choose_crayons(
            item=poly_item,
            border_style=border_style,
            fill=fill)

        # Add polygon item to layer
        layer.Polygons.append(poly_item)

    @classmethod
    def add_open(cls, layer: 'Layer', asset: str, vertices: List[Position]):
        """
        Rather than manage a seperate render open polygon list in the Layer, we
        can just add these to the line segment list.

        :param layer: Draw on this layer
        :param asset: Used to look up the line style
        :param vertices: A sequences of 2 or more vertices
        """
        for v1, v2 in zip(vertices, vertices[1:]):
            assert len(vertices) > 1, "Open polygon has less than two vertices"
            LineSegment.add(layer=layer, asset=asset, from_here=v1, to_there=v2)

    @classmethod
    def render(cls, layer: 'Layer'):
        """
        Draw the closed non-rectangular shapes

        :param layer: Draw on this layer
        """
        for p in layer.Polygons:
            # _logger.info(f"> Polygon at: [{p.vertices}]")

            layer.Scene.addItem(p)


"""
tabletqt.py – A multi-layered drawing surface implemented on top of the Qt GUI framework
"""
# System
import logging
from datetime import datetime  # For initial log entry
from pathlib import Path
from typing import Optional

# Qt
from PyQt6.QtWidgets import QApplication

# Tablet
from tabletqt.exceptions import NonSystemInitialLayer, TabletBoundsExceeded, MissingConfigData
from tabletqt.geometry_types import Rect_Size, Position
from tabletqt.styledb import StyleDB
from tabletqt.graphics.text_element import TextElement
from tabletqt.layer import Layer
from tabletqt.scene_view import MainWindow
from tabletqt.graphics.image import ImageDE
from tabletqt.configuration.styles import FloatRGB
from tabletqt.graphics.symbol import Symbol

default_background = FloatRGB(255, 255, 255)  # White

gui_app = QApplication([]) # Declare it here so it is not garbage collected on repeated instantiation of Tablet
# We create a Qt App instance just once on the import and then keep using the same one

class Tablet:
    """
    The Tablet class is part of the Drawing domain which provides a service to an application
    with simple 2D diagramming needs such as the Flatland model diagram generator.

    Most importantly, the Drawing domain and hence the Tablet abstract away the details of
    graphics library interaction from the fundamental diagramming application.

    The original implementation of Tablet supported the Cairo graphics library while the current
    implementation is built on the Qt Gui framework. No changes are necessary in any client diagramming
    applications that draw on a Tablet.

    For example, when a Flatland node compartment wants to draw itself, it doesn't worry about line
    widths, dash patterns and colors. It also doesn't worry about flipping to whatever coordinate
    system the graphics library uses. A compartment would just say "add_shape( asset, size, pin )" and
    the Tablet will take care of the rest. Here the asset is the name of the model entity 'compartment',
    the size in points (or whatever units the Flatland app wants to use) and pin expressed in Flatland
    Canvas coordinates.

    When a Tablet is created, it is initialized with a single populated Layer where Assets can be drawn.
    For example, Flatland might start with a predefined 'diagram' Layer using a Drawing Type such as
    'state machine diagram' and a Presentation such as 'default' or 'formal' or 'corporate'.

    The Diagram Type determines what kinds of things might be drawn, Assets such as 'connector',
    'compartment', 'class name', etc., while the Presentation establishes the Text and Line Styles to
    be used when drawing those Assets. All of this information is stored in yaml files which the Tablet
    loads upon creation.

    Each time the application wants something drawn, it will ask Tablet to add the appropriate Asset
    to one of its render lists. A separate list is maintained for all the rectangles, line segments and
    text lines to be rendered. When the application is finished populating these lists, the Tablet can
    render everything using its graphic library (Qt in this implementation) using whatever coordinate
    system the library supplies, making any necessary conversions from the application coordinate system.

        Attributes and relationships defined on the class model

        - ID {I} -- Unique id, implemented as a reference to a Tablet object, no attribute needed
        - Size -- The height and width of the drawing surface (attribute)
        - Output_file -- A filename or output stream object to be output as a drawing (attribute)
        - Background_color -- The color of the Tablet (visible through all non-opaque layer elements)
    """

    def __init__(self, size: Rect_Size, output_file: Path, drawing_type: str, presentation: str,
                 layer: str, show_window: bool = False, background_color: Optional[str] = 'white'):
        """
        Constructs a new Tablet instance with a single initial predefined Layer

        :param size: Vertical and horizontal span of the entire draw surface in points
        :param output_file: Name of the drawing file to be generated, PDF only for now
        :param drawing_type: Initial layer Drawing Type so we know what kinds text and graphics Assets can be drawn
        :param presentation: Initial layer's Presentation so we know what graphic styles to use for our Assets
        :param layer: The name of the predefined initial Layer to be created on this Tablet (typically 'diagram')
        :param background_color: Name of background color defined in colors.yaml
        """
        self.logger = logging.getLogger(__name__)
        self.logger.info(f"Tablet init: {datetime.now().strftime('%Y-%m-%d %H:%M:%S')}\n")
        self.app_name = "tablet"

        # Load all of the common font, color, etc. styles used by all Presentations from yaml files
        StyleDB.load_config_files()

        # Load symbols, stickers, and build image paths for all assets and drawing types
        Symbol.load_symbol_defs()
        ImageDE.build_paths()
        TextElement.load_stickers()

        # Establish a system default layer ordering. Not all of them will be used in any given
        # View, but this is the draw order from bottom-most layer upward
        # It can (should be) customizable by the user, but this should work for most diagrams
        self.show_window = show_window
        try:
            self.background_color = StyleDB.color[background_color]  # This is referenced when filling text underlay rects
        except KeyError:
            self.logger.error(f'Background color [{background_color} not defined in colors.yaml file')
            raise MissingConfigData
        self.layer_order = ['sheet', 'grid', 'frame', 'diagram', 'scenario', 'annotation']
        self.Presentations = {}  # Presentations loaded from the Flatland database, updated by Layer class
        self.App = gui_app  # QT Application (must be created before any QT widgets)
        self.Window = MainWindow(title=self.app_name, size=size, background=self.background_color)  # QT widget for drawing 2D elements
        self.View = self.Window.graphics_view

        if layer not in self.layer_order:
            raise NonSystemInitialLayer
        self.layers = {layer: Layer(name=layer, tablet=self, presentation=presentation, drawing_type=drawing_type)}
        # Initialize the first layer at the indicated position. If the position is not in the system layer order
        # list, it will be placed as the topmost layer. Usually, though, the initial layer should be diagram

        # self.Drawing_type = drawing_type  # class diagram, state diagram, etc
        self.Size = size
        self.Output_file = output_file

    def add_layer(self, name: str, presentation: str, drawing_type: str) -> Optional[Layer]:
        """
        Populate a new layer by name and return it. If a layer of the same name has already been
        populated, no layer is returned.

        All drawing takes place on a designated layer, such as 'diagram' or 'sheet'.

        A defined layer is simply a known layer name within a predefined rendering order.
        You can think of each position in the order as a z coordinate with 0 corresponding to the first
        rendered layer. For example, the 'sheet' layer is at the bottom (0) with 'diagram' somewhere
        above it.

        A layer, predefined or custom, must be populated before anything can be drawn on it.
        The client application is responsible for populating any layers that it needs. There is not automatic
        population.

        If a layer name is supplied that does not correspond to any of the predefined layers, it will be stacked
        after the last predefined layer and, thus, rendered last.

        :param name: One of the standard layer stickers or a custom layer name
        :param presentation: The Presentation name associated with this Layer
        :param drawing_type: The Drawing Type defining this Presentation
        :return: A reference to the newly created layer
        """
        if not self.layers.get(name):
            if name not in self.layer_order:
                self.layer_order.append(name)
            self.layers[name] = Layer(name=name, tablet=self, presentation=presentation, drawing_type=drawing_type)
            return self.layers[name]
        else:
            self.logger.warning(f"Layer: [{name}] already exists")
            return None

    def render(self):
        """
        Renders each populated layer of the Tablet moving up the z axis. Any unpopulated layers are skipped.
        """
        # Create and show the drawing window
        [self.layers[name].render() for name in self.layer_order if self.layers.get(name)]
        self.Window.show()

        # Save the rendered tabletqt as a PDF for alternate viewing
        self.View.save_as_pdf(self.Output_file)

        # Run the Qt GUI event loop
        # sys.exit(self.App.exec())
        if self.show_window:
            self.App.exec()

    def to_dc(self, tablet_coord: Position) -> Position:
        """
        Convert from tabletqt coordinates (tc) used by the client application to device
        coordinates (dc).

        Tablet coordinates are upper right quadrant cartesian with the origin (0,0) in the
        lower left corner of the tabletqt. This is what client application (user) specifies
        when drawing.

        Device coordinates depend on the graphics library. Here we are using Qt, so we have
        a standard display coordinate system with the origin in the upper left corner with
        y values ascending toward the bottom of the display.

        An exception is thrown if the supplied position is outside of the tabletqt boundary.

        Note: This may seem like overkill for a simple computation and check, but less
        error prone than having this pattern sprinked throughout the code.

        :param tablet_coord: Position in table coordinates
        :return: Position in device coordinates
        """
        if tablet_coord.y > self.Size.height:
            raise TabletBoundsExceeded("Tablet height exceeded", height=tablet_coord.y - self.Size.height, width=0)
        assert tablet_coord.x >= 0, "Negative x value"
        assert tablet_coord.y >= 0, "Negative y value"
        return Position(x=tablet_coord.x, y=self.Size.height - tablet_coord.y)

    def __repr__(self):
        return f'Size: {self.Size},  Output: {self.Output_file}'



"""
styledb.py - Loads styles common to all Presentations
"""
# System
import logging
from pathlib import Path
from typing import NamedTuple, Any

# Modelint
from mi_config.config import Config

# Tablet
from tabletqt.exceptions import BadConfigData
from tabletqt.configuration.styles import (FloatRGB, LineStyle, TextStyle, DashPattern)
from tabletqt.tablet_config import TabletConfig

_logger = logging.getLogger(__name__)

# nt - named tuple is defined
# pre/post - Whether or not the data must be pre or post processed
# PP = namedtuple('PP', 'nt pre post')  # Postprocess configuration file data

class PP(NamedTuple):
    nt: Any
    pre: bool
    post: bool


# yaml file name : load data using this tuple
config_type = {
    'colors': PP(nt=FloatRGB, pre=True, post=False),
    'line_styles': PP(nt=LineStyle, pre=False, post=False ),
    'dash_patterns': PP(nt=DashPattern, pre=False, post=False),
    'typefaces': PP(nt=None, pre=False, post=False),
    'text_styles': PP(nt=TextStyle, pre=False, post=True),
    'color_usages': PP(nt=None, pre=False, post=True),
    'symbols': PP(nt=None, pre=False, post=False),
}


class StyleDB:
    """
    Manages a set of graphic and text styles defined for the Drawing Types that will be used
    with a Tablet. The client app can then refer to these styles rather than explicitly specifying
    graphics properties like font, line width, color fill and so forth.

    For example, the client app might request a 'state activity compartment' and then obtain
    all the predefined styles automatically.

    The styles are all defined in a set of yaml files that reference each other's
    elements.
    """
    styles = {}  # { 'text_style':
    color = {}  # rgb color float representation
    typeface = None
    dash_pattern = None
    line_style = None
    text_style = None
    color_usage = None
    raw_config_data = None
    config_data = None
    symbol = None


    @classmethod
    def load_config_files(cls):
        """
        Processes the config_type dictionary, loading each yaml configuration file into either
        a named tuple or a simple key value dictionary if no named tuple is provided
        and then sets the corresponding StyleDB class attribute to that value
        """
        _logger.info(f"StyleDB loading tabletqt configuration\n---")
        fspec = {k: v.nt for k,v in config_type.items()}
        cls.raw_config_data = Config(app_name=TabletConfig.app_name, lib_config_dir=TabletConfig.config_path,
                                     fspec=fspec)
        cls.config_data = cls.raw_config_data.loaded_data

        for fname, cdata in cls.config_data.items():
            config_file_path = TabletConfig.config_path / (fname + ".yaml")
            _logger.info(f"loading: {config_file_path}")
            if config_type[fname].pre:
                # Retrieve the relevant preprocessing method name using the filename suffix
                method_name = 'preprocess_'+fname
                method = getattr(cls, method_name, None)
                # Invoke it on the loaded yaml data
                method(cls.config_data[fname])
            attr_name = fname[:-1]  # drop the plural 's' from the file name to get the attribute name
            # Assign the loaded and possibly preprocessed yaml data to the relevant class attribute
            setattr(cls, attr_name, cls.config_data[fname])
            if config_type[fname].post:
                method_name = 'postprocess_'+fname  # Keep the plural
                method = getattr(cls, method_name, None)
                method()

        _logger.info(f"---\n")

    @classmethod
    def postprocess_text_styles(cls):
        """
        Verify all typefaces are defined
        """
        undefined_typefaces = [t.typeface for t in cls.text_style.values() if t.typeface not in cls.typeface]
        if undefined_typefaces:
            _logger.error(f"Undefined typefaces: {undefined_typefaces} encountered in"
                          f"text styles configuration file:\n    {config_path / 'text_styles.yaml'}")
            raise BadConfigData


    @classmethod
    def preprocess_colors(cls, raw_data):
        """
        # Validate color float values for QT6
        """
        for name, rgb in raw_data.items():
            for n in [rgb.r, rgb.g, rgb.b]:
                if not 0 <= n <= 255:
                    _logger.error(f"Bad color value [{n}] for: {name} in "
                                  f"configuration file:\n    {config_path / 'colors.yaml'}")
                    raise BadConfigData
            StyleDB.color[name] = rgb

    @classmethod
    def postprocess_color_usages(cls):
        """
        Validate color stickers
        """
        undefined_colors = [c for c in cls.color_usage.values() if c not in cls.color]
        if undefined_colors:
            _logger.error(f"Undefined colors: {undefined_colors} encountered in"
                          f"color usages configuration file:\n    {config_path / 'color_usages.yaml'}")
            raise BadConfigData

    @classmethod
    def report_colors(cls):
        """
        Prints out a list of the available colors for the user
        """
        cls.load_config_files()
        print("Canvas colors:")
        print("---")
        for c in cls.rgbF:
            print(c)
        print("===")



""" diagnostic_marker.py -- Diagnostic marking """

from tabletqt.geometry_types import Position, Rect_Size

from PyQt6.QtWidgets import QGraphicsLineItem, QGraphicsRectItem
from PyQt6.QtCore import QRectF, QLineF
from typing import TYPE_CHECKING
from PyQt6.QtGui import QPen, QColor

if TYPE_CHECKING:
    from tabletqt.layer import Layer


class DiagnosticMarker:
    """
    Manage the rendering of diagnostic markers such as crosshairs, rectangles
    and other shapes drawn with explicitly specified styles.

    (The StyleDB is not consulted for any of these)
    """
    ch_radius = 3  # Cross hair radius

    @classmethod
    def add_cross_hair(cls, layer: 'Layer', location: Position, color: str = 'purple'):
        """
        Place a diagnostic cross hair at the requested point in the Scene.
        This method is not intended for client application use.

        :param color:
        :param location: Where to place the crosshair in application coordinates
        """
        pen = QPen(QColor(color), 1)
        # Flip to device coordinates
        ll_dc = layer.Tablet.to_dc(Position(x=location.x, y=location.y))
        hline = QGraphicsLineItem(QLineF(ll_dc.x-cls.ch_radius, ll_dc.y, ll_dc.x+cls.ch_radius, ll_dc.y))
        hline.setPen(pen)
        vline = QGraphicsLineItem(QLineF(ll_dc.x, ll_dc.y-cls.ch_radius, ll_dc.x, ll_dc.y+cls.ch_radius))
        vline.setPen(pen)
        layer.RawLines.append(hline)
        layer.RawLines.append(vline)

    @classmethod
    def add_raw_rectangle(cls, layer: 'Layer', upper_left: Position, size: Rect_Size):
        """
        Adds an unfilled rectangle for diagnostic purposes. 'raw' means that the dimensions
        are supplied directly without consulting the StyleDB

        :param upper_left:
        :param size:
        :return:
        """
        # Use upper left corner instead
        # ul = Position(x=upper_left.x, y=upper_left.y - size.height)

        # Convert upper left corner to device coordinates
        uldc = layer.Tablet.to_dc(Position(x=upper_left.x, y=upper_left.y))

        rect = QRectF(uldc.x, uldc.y, size.width, size.height)
        rect_item = QGraphicsRectItem(rect)
        layer.RawRectangles.append(rect_item)

    @classmethod
    def render(cls, layer: 'Layer'):

        # Render all raw lines
        for l in layer.RawLines:
            layer.Scene.addItem(l)

        # Render all raw rectangles
        pen = QPen(QColor(0, 0, 0))
        for r in layer.RawRectangles:
            r.setPen(pen)
            layer.Scene.addItem(r)



"""
presentation.py – Presentation class in View domain
"""
# System
import logging
from collections import namedtuple
from pathlib import Path

# Modelint
from mi_config.config import Config

CornerSpec = namedtuple('Corner_Spec', 'radius top bottom')
config_dir = Path(__file__).parent / "configuration"

_logger = logging.getLogger(__name__)


class Presentation:
    """
   A set of compatible visual styles including fonts, colors, border widths and so forth as appropriate to a
   given View Type form a selectable Presentation. For example, an `Executable UML State Machine Diagram`
   might be drawn using certain fonts for state stickers and possibly different colors for transient and
   non-transient states. Alternatively, only black and white might be used with purple for a certain kind of
   connector in a diagnostic Presentation.
   """

    def __init__(self, name: str, drawing_type: str):
        """
       Constructor
       """
        self.logger = logging.getLogger(__name__)
        self.Name = name
        self.Drawing_type = drawing_type
        self.Text_presentation = {}
        self.Underlays = None  # Set of text presentations that require an underlay
        self.Line_presentation = {}
        self.Rectangle_presentation = {}
        self.Symbol_presentation = {}
        self.Corner_spec = {}

        # Load Asset Presentations for all Assets in this Presentation
        self.logger.info(f"using presentation: [{self.Name}]")
        self.load_drawing_type()

    def load_drawing_type(self):
        """
        Load symbol, shape, and text assets for my drawing type and presentation
        """
        _logger.info(f"Loading presentations\n---")

        c = Config(app_name='mi_tablet', lib_config_dir=config_dir, fspec={'drawing_types':None})
        dtype_data = c.loaded_data['drawing_types']
        try:
            my_data = dtype_data[self.Drawing_type][self.Name]
        except KeyError:
            _logger.error(f"No presentation [{self.Name}] defined for drawing type [{self.Drawing_type}]")
            raise

        # Load symbol presentations
        self.Symbol_presentation = my_data.get('symbol')
        self.Text_presentation = my_data.get('text')
        self.Underlays = {k for k,v in self.Text_presentation.items() if v.get('underlay')}
        if my_data.get('shape'):
            self.Rectangle_presentation = my_data['shape'].get('rectangle')
            self.Line_presentation = my_data['shape'].get('line')


""" circle_se.py -- Circle shape element """

# System
import logging
from typing import TYPE_CHECKING

# Qt
from PyQt6.QtWidgets import QGraphicsEllipseItem
from PyQt6.QtCore import QRectF

# Tablet
import tabletqt.element as element
from tabletqt.geometry_types import Position
from tabletqt.graphics.crayon_box import CrayonBox

if TYPE_CHECKING:
    from tabletqt.layer import Layer

_logger = logging.getLogger(__name__)
class CircleSE:
    """
    Manage the rendering of Circle Shape Elements

    Attributes and relationships defined on the class model

    Subclass of Closed Shape on class model (R22)

    - ID {I} -- Element ID, unique within a Layer, implemented as object reference
    - Layer {I, R22} -- Element drawn on this Layer via R22/R12/R15/Element/R19/Layer
    - Radius -- Circle radius
    - Center -- Position of center in tablet coordinates
    """

    @classmethod
    def add(cls, layer: 'Layer', asset: str, center: Position, radius: float):
        """
        Adds a circle to the layer and converts the center to device coordinates

        :param layer: Draw on this layer
        :param asset: Used to look up the border style
        :param center: The center position in tablet coordinates
        :param radius: The radius length in points
        """
        # Flip lower left corner to device coordinates
        center_dc = layer.Tablet.to_dc(Position(x=center.x, y=center.y))

        # Check to see if this circle is filled
        fill = layer.Presentation.Closed_shape_fill.get(asset)

        layer.Circles.append(element.Circle(
            center=center_dc, radius=radius, border_style=layer.Presentation.Shape_presentation[asset], fill=fill,
        ))

    @classmethod
    def render(cls, layer: 'Layer'):
        """
        Draw the circle shapes

        :param layer: Draw on this layer
        """

        for c in layer.Circles:

            # Create the circle item
            diameter = c.radius*2
            c_item = QGraphicsEllipseItem(QRectF(c.center.x, c.center.y, diameter, diameter))
            _logger.info(f"> Circle at: ({c.center.x}, {c.center.y}), diameter: {diameter}")

            # Set pen and brush
            CrayonBox.choose_crayons(item=c_item, border_style=c.border_style, fill=c.fill)

            # Add it to the scene
            layer.Scene.addItem(c_item)




""" demo6_bp_text.py - Test rectangles using blueprint presentation """

from pathlib import Path
from tabletqt.tablet import Tablet
from tabletqt.geometry_types import Rect_Size, Position
from tabletqt.graphics.text_element import TextElement, TextBlockCorner, HorizAlign
from tabletqt.graphics.line_segment import LineSegment

points_in_mm = 2.83465


class SketchSymbols:
    tablet = None
    layer = None
    size = None

    @classmethod
    def make_a_tablet(cls):
        cls.size = Rect_Size(round(210 * points_in_mm), round(297 * points_in_mm))  # Ansi C portrait
        dtype = "xUML class diagram"
        # dtype = "Starr class diagram"
        # dtype = "xUML state machine diagram"
        output_path = Path(__file__).parent.parent.parent.parent / "working" / "demo9_underlay.pdf"
        cls.tablet = Tablet(size=cls.size, output_file=output_path, drawing_type=dtype,
                            presentation="default", layer="diagram", show_window=False, background_color='blue steel')

    @classmethod
    def sketch(cls):
        cls.make_a_tablet()
        dlayer = cls.tablet.layers['diagram']

        # LineSegment.add(layer=dlayer, asset="binary association connector",
        #                 from_here=Position(100,305), to_there=Position(400,305))

        LineSegment.add(layer=dlayer, asset="binary association connector",
                        from_here=Position(100,305), to_there=Position(400,305))

        TextElement.add_sticker(layer=dlayer, asset='superclass face name', name='superclass',
                                pin=Position(300, 300), corner=TextBlockCorner.LL)
        # TextElement.add_block(layer=dlayer, asset='superclass face name', lower_left=Position(300, 300), text=['underlay'])

        cls.tablet.render()


if __name__ == "__main__":
    SketchSymbols.sketch()



"""
__main__.py
"""

# System
import logging
import logging.config
import sys
import argparse
from pathlib import Path

# Tablet
from tabletqt import version
# from tabletqt.etchasketch import EtchaSketch
# from tabletqt.sketchsymbols import SketchSymbols

_logpath = Path("tabletqt.log")
app_name = "Tablet"

def get_logger():
    """Initiate the logger"""
    log_conf_path = Path(__file__).parent / 'log.conf'  # Logging configuration is in this file
    logging.config.fileConfig(fname=log_conf_path, disable_existing_loggers=False)
    return logging.getLogger(__name__)  # Create a logger for this module

# Configure the expected parameters and actions for the argparse module
def parse(cl_input):
    """
    The command line interface is for diagnostic purposes

    :param cl_input:
    :return:
    """
    parser = argparse.ArgumentParser(description='Tabletx 2D draw interface to Cairo')
    parser.add_argument('-demo', action='store',
                        help='name of demo module')
    parser.add_argument('-COLORS', '--colors', action='store_true',
                        help='Show the list of background color stickers')
    parser.add_argument('-T', '--test', action='store_true',
                        help='Run some test features')
    parser.add_argument('-D', '--debug', action='store_true',
                        help='Debug mode'),
    parser.add_argument('-V', '--version', action='store_true',
                        help='Print the current version')
    return parser.parse_args(cl_input)


def main():
    # Start logging
    logger = get_logger()
    logger.info(f'{app_name} version: {version}')
    logger.info('-----------------\n')

    # Parse the command line args
    args = parse(sys.argv[1:])

    if args.version:
        # Just print the version and quit
        print(f'{app_name} version: {version}')
        sys.exit(0)

    if args.demo:
        spath = Path(__file__).parent / f'demo/{args.demo}.py'
        sys.exit(0)

    if args.test:
        from tabletqt.styledb import StyleDB
        StyleDB.load_config_files()

    if args.colors:
        # Just print the database colors and quit
        from tabletqt.styledb import StyleDB
        StyleDB.report_colors()

    logger.info("No problemo")  # We didn't die on an exception
    print("\nNo problemo")


if __name__ == "__main__":
    main()